import pytest

from textacy.similarity import edits


@pytest.mark.parametrize(
    "str1, str2, exp",
    [
        ("abcd", "abcd", 1.0),
        ("abdc", "wxyz", 0.0),
        ("abcd", "", 0.0),
        ("", "", 0.0),
        ("abcd", "abcD", 0.75),
        ("Abcd", "abcd", 0.75),
        ("abcd", "abcdefgh", 0.5),
        ("abcdefgh", "abcd", 0.5),
    ]
)
def test_hamming(str1, str2, exp):
    obs = edits.hamming(str1, str2)
    assert isinstance(obs, float)
    assert 0.0 <= obs <= 1.0
    assert obs == pytest.approx(exp, rel=0.01)


@pytest.mark.parametrize(
    "str1, str2, exp",
    [
        ("abcd", "abcd", 1.0),
        ("abdc", "wxyz", 0.0),
        ("abcd", "", 0.0),
        ("", "", 0.0),
        ("abcd", "abcD", 0.75),
        ("Abcd", "abcd", 0.75),
        ("abcd", "bacd", 0.5),
        ("abcd", "abXcd", 0.8),
        ("abcd", "abcdefgh", 0.5),
        ("abcdefgh", "abcd", 0.5),
    ]
)
def test_levenshtein(str1, str2, exp):
    obs = edits.levenshtein(str1, str2)
    assert isinstance(obs, float)
    assert 0.0 <= obs <= 1.0
    assert obs == pytest.approx(exp, rel=0.01)


@pytest.mark.parametrize(
    "str1, str2, exp",
    [
        ("abcd", "abcd", 1.0),
        ("abdc", "wxyz", 0.0),
        ("abcd", "", 0.0),
        ("", "", 0.0),
        ("abcd", "abcD", 0.8333),
        ("Abcd", "abcd", 0.8333),
        ("abcd", "abcdefgh", 0.8333),
        ("abcdefgh", "abcd", 0.8333),
    ]
)
def test_jaro(str1, str2, exp):
    obs = edits.jaro(str1, str2)
    assert isinstance(obs, float)
    assert 0.0 <= obs <= 1.0
    assert obs == pytest.approx(exp, rel=0.01)



"""
KWIC
----

:mod:`textacy.extract.kwic`: Extract keywords with their surrounding contexts from
a text document using regular expressions.
"""
from __future__ import annotations

import re
from typing import Iterable, Pattern, Tuple

from spacy.tokens import Doc


def keyword_in_context(
    doc: Doc | str,
    keyword: str | Pattern,
    *,
    ignore_case: bool = True,
    window_width: int = 50,
    pad_context: bool = False,
) -> Iterable[Tuple[str, str, str]]:
    """
    Search for ``keyword`` matches in ``doc`` via regular expression and yield matches
    along with ``window_width`` characters of context before and after occurrence.

    Args:
        doc: spaCy ``Doc`` or raw text in which to search for ``keyword``. If a ``Doc``,
            constituent text is grabbed via :attr:`spacy.tokens.Doc.text`. Note that
            spaCy annotations aren't used at all here, they're just a convenient
            owner of document text.
        keyword: String or regular expression pattern defining the keyword(s) to match.
            Typically, this is a single word or short phrase ("spam", "spam and eggs"),
            but to account for variations, use regex (``r"[Ss]pam (and|&) [Ee]ggs?"``),
            optionally compiled (``re.compile(r"[Ss]pam (and|&) [Ee]ggs?")``).
        ignore_case: If True, ignore letter case in ``keyword`` matching; otherwise,
            use case-sensitive matching. Note that this argument is only used if
            ``keyword`` is a string; for pre-compiled regular expressions,
            the ``re.IGNORECASE`` flag is left as-is.
        window_width: Number of characters on either side of ``keyword``
            to include as "context".
        pad_context: If True, pad pre- and post-context strings to ``window_width``
            chars in length; otherwise, us as many chars as are found in the text,
            up to the specified width.

    Yields:
        Next matching triple of (pre-context, keyword match, post-context).
    """
    text = doc.text if isinstance(doc, Doc) else doc
    if isinstance(keyword, str):
        flags = re.IGNORECASE if ignore_case is True else 0
        matches = re.finditer(keyword, text, flags=flags)
    else:
        matches = keyword.finditer(text)
    for match in matches:
        pre_context = text[max(0, match.start() - window_width) : match.start()]
        post_context = text[match.end() : match.end() + window_width]
        if pad_context is True:
            pre_context = pre_context.rjust(window_width)
            post_context = post_context.ljust(window_width)

        yield (pre_context, match.group(), post_context)



import pytest

from textacy import preprocessing


@pytest.mark.parametrize(
    "text_in, fast, text_out",
    [
        (
            "El niño se asustó del pingüino -- qué miedo!",
            True,
            "El nino se asusto del pinguino -- que miedo!",
        ),
        (
            "El niño se asustó del pingüino -- qué miedo!",
            False,
            "El nino se asusto del pinguino -- que miedo!",
        ),
        (
            "Le garçon est très excité pour la forêt.",
            True,
            "Le garcon est tres excite pour la foret.",
        ),
        (
            "Le garçon est très excité pour la forêt.",
            False,
            "Le garcon est tres excite pour la foret.",
        ),
    ]
)
def test_remove_accents(text_in, fast, text_out):
    assert preprocessing.remove.accents(text_in, fast=fast) == text_out


@pytest.mark.parametrize(
    "text_in, only, text_out",
    [
        ("Hello, {name}!", None, "Hello, !"),
        ("Hello, world (DeWilde et al., 2021, p. 42)!", None, "Hello, world !"),
        ("Hello, world (1)!", None, "Hello, world !"),
        ("Hello, world [1]!", None, "Hello, world !"),
        (
            "Hello, world (and whomever it may concern [not that it's any of my business])!",
            None,
            "Hello, world !",
        ),
        (
            "Hello, world (and whomever it may concern (not that it's any of my business))!",
            None,
            "Hello, world (and whomever it may concern )!",
        ),
        (
            "Hello, world (and whomever it may concern [not that it's any of my business])!",
            "square",
            "Hello, world (and whomever it may concern )!",
        ),
        ("Hello, world [1]!", "round", "Hello, world [1]!"),
        ("Hello, world [1]!", ("curly", "round"), "Hello, world [1]!"),
    ]
)
def test_remove_brackets(text_in, only, text_out):
    assert preprocessing.remove.brackets(text_in, only=only) == text_out


@pytest.mark.parametrize(
    "text_in, text_out",
    [
        ("Hello, <i>world!</i>", "Hello, world!"),
        ("<title>Hello, world!</title>", "Hello, world!"),
        ('<title class="foo">Hello, world!</title>', "Hello, world!"),
        (
            "<html><head><title>Hello, <i>world!</i></title></head></html>",
            "Hello, world!",
        ),
        (
            "<html>\n"
            "  <head>\n"
            '    <title class="foo">Hello, <i>world!</i></title>\n'
            "  </head>\n"
            "  <!--this is a comment-->\n"
            "  <body>\n"
            "    <p>How's it going?</p>\n"
            "  </body>\n"
            "</html>",
            "Hello, world!\n  \n  \n  \n    How's it going?",
        ),
    ]
)
def test_remove_html_tags(text_in, text_out):
    assert preprocessing.remove.html_tags(text_in) == text_out


@pytest.mark.parametrize(
    "text_in, only, text_out",
    [
        (
            "I can't. No, I won't! It's a matter of \"principle\"; of -- what's the word? -- conscience.",
            None,
            "I can t  No  I won t  It s a matter of  principle   of    what s the word     conscience ",
        ),
        (
            "I can't. No, I won't! It's a matter of \"principle\"; of -- what's the word? -- conscience.",
            ".",
            "I can't  No, I won't! It's a matter of \"principle\"; of -- what's the word? -- conscience ",
        ),
        (
            "I can't. No, I won't! It's a matter of \"principle\"; of -- what's the word? -- conscience.",
            ["-", "'", "\""],
            "I can t. No, I won t! It s a matter of  principle ; of   what s the word?   conscience.",
        ),
    ]
)
def test_remove_punct(text_in, only, text_out):
    assert preprocessing.remove.punctuation(text_in, only=only) == text_out



"""
Basic Stats
-----------

:mod:`textacy.text_stats.basics`: Low-level functions for computing basic text statistics,
typically accessed via :class:`textacy.text_stats.TextStats`.
"""
import logging
import math
from typing import Iterable, Tuple, Union

import spacy.pipeline
from cytoolz import itertoolz
from spacy.tokens import Doc, Token

from .. import extract
from . import api as _api


LOGGER = logging.getLogger(__name__)

_SENTENCIZER = spacy.pipeline.Sentencizer()


def n_words(doc_or_words: Union[Doc, Iterable[Token]]) -> int:
    """
    Compute the number of words in a document.

    Args:
        doc_or_words: If a spaCy ``Doc``, non-punctuation tokens (words) are extracted;
            if an iterable of spaCy ``Token`` s, all are included as-is.
    """
    words = _get_words(doc_or_words)
    return itertoolz.count(words)


def n_unique_words(doc_or_words: Union[Doc, Iterable[Token]]) -> int:
    """
    Compute the number of *unique* words in a document.

    Args:
        doc_or_words: If a spaCy ``Doc``, non-punctuation tokens (words) are extracted;
            if an iterable of spaCy ``Token`` s, all are included as-is.
    """
    words = _get_words(doc_or_words)
    # NOTE: this stdlib solution is slower than itertoolz for docs with ~250+ words
    # so let's take a small hit on short docs for the sake of big wins on long docs
    # return len({word.lower for word in words})
    return itertoolz.count(itertoolz.unique(word.lower for word in words))


def n_chars_per_word(doc_or_words: Union[Doc, Iterable[Token]]) -> Tuple[int, ...]:
    """
    Compute the number of characters for each word in a document.

    Args:
        doc_or_words: If a spaCy ``Doc``, non-punctuation tokens (words) are extracted;
            if an iterable of spaCy ``Token`` s, all are included as-is.
    """
    words = _get_words(doc_or_words)
    return tuple(len(word) for word in words)


def n_chars(n_chars_per_word: Tuple[int, ...]) -> int:
    """
    Compute the total number of characters in a document.

    Args:
        n_chars_per_word: Number of characters per word in a given document,
            as computed by :func:`n_chars_per_word()`.
    """
    return sum(n_chars_per_word)


def n_long_words(n_chars_per_word: Tuple[int, ...], min_n_chars: int = 7) -> int:
    """
    Compute the number of long words in a document.

    Args:
        n_chars_per_word: Number of characters per word in a given document,
            as computed by :func:`n_chars_per_word()`.
        min_n_chars: Minimum number of characters required for a word to be
            considered "long".
    """
    return itertoolz.count(nc for nc in n_chars_per_word if nc >= min_n_chars)


def n_syllables_per_word(
    doc_or_words: Union[Doc, Iterable[Token]], lang: str,
) -> Tuple[int, ...]:
    """
    Compute the number of syllables for each word in a document.

    Args:
        doc_or_words: If a spaCy ``Doc``, non-punctuation tokens (words) are extracted;
            if an iterable of spaCy ``Token`` s, all are included as-is.

    Note:
        Identifying syllables is _tricky_; this method relies on hyphenation, which is
        more straightforward but doesn't always give the correct number of syllables.
        While all hyphenation points fall on syllable divisions, not all syllable
        divisions are valid hyphenation points.
    """
    hyphenator = _api.load_hyphenator(lang=lang)
    words = _get_words(doc_or_words)
    return tuple(len(hyphenator.positions(word.lower_)) + 1 for word in words)


def n_syllables(n_syllables_per_word: Tuple[int, ...]) -> int:
    """
    Compute the total number of syllables in a document.

    Args:
        n_syllables_per_word: Number of syllables per word in a given document,
            as computed by :func:`n_syllables_per_word()`.
    """
    return sum(n_syllables_per_word)


def n_monosyllable_words(n_syllables_per_word: Tuple[int, ...]) -> int:
    """
    Compute the number of monosyllobic words in a document.

    Args:
        n_syllables_per_word: Number of syllables per word in a given document,
            as computed by :func:`n_syllables_per_word()`.
    """
    return itertoolz.count(ns for ns in n_syllables_per_word if ns == 1)


def n_polysyllable_words(
    n_syllables_per_word: Tuple[int, ...], min_n_syllables: int = 3,
) -> int:
    """
    Compute the number of polysyllobic words in a document.

    Args:
        n_syllables_per_word: Number of syllables per word in a given document,
            as computed by :func:`n_syllables_per_word()`.
        min_n_syllables: Minimum number of syllables required for a word to be
            considered "polysyllobic".
    """
    return itertoolz.count(ns for ns in n_syllables_per_word if ns >= min_n_syllables)


def _get_words(doc_or_words: Union[Doc, Iterable[Token]]) -> Iterable[Token]:
    if isinstance(doc_or_words, Doc):
        return extract.words(
            doc_or_words, filter_punct=True, filter_stops=False, filter_nums=False,
        )
    else:
        return doc_or_words


def n_sents(doc: Doc) -> int:
    """
    Compute the number of sentences in a document.

    Warning:
        If ``doc`` has not been segmented into sentences, it will be modified in-place
        using spaCy's rule-based ``Sentencizer`` pipeline component before counting.
    """
    if not doc.has_annotation("SENT_START"):
        LOGGER.warning(
            "`doc` has not been segmented into sentences; applying spaCy's rule-based, "
            "`Sentencizer` pipeline component to `doc` before counting..."
        )
        doc = _SENTENCIZER(doc)
    return itertoolz.count(doc.sents)


def entropy(doc_or_words: Union[Doc, Iterable[Token]]) -> float:
    """
    Compute the entropy of words in a document.

    Args:
        doc_or_words: If a spaCy ``Doc``, non-punctuation tokens (words) are extracted;
            if an iterable of spaCy ``Token`` s, all are included as-is.
    """
    words = _get_words(doc_or_words)
    word_counts = itertoolz.frequencies(word.text for word in words)
    n_words = sum(word_counts.values())
    probs = (count / n_words for count in word_counts.values())
    return -sum(prob * math.log2(prob) for prob in probs)



"""
Replace
-------

:mod:`textacy.preprocessing.replace`: Replace parts of raw text that are semantically
important as members of a group but not so much in the individual instances. Can also
be used to remove such parts by specifying ``repl=""`` in function calls.
"""
from . import resources


def currency_symbols(text: str, repl: str = "_CUR_") -> str:
    """Replace all currency symbols in ``text`` with ``repl``."""
    return resources.RE_CURRENCY_SYMBOL.sub(repl, text)


def emails(text: str, repl: str = "_EMAIL_") -> str:
    """Replace all email addresses in ``text`` with ``repl``."""
    return resources.RE_EMAIL.sub(repl, text)


def emojis(text: str, repl: str = "_EMOJI_") -> str:
    """
    Replace all emoji and pictographs in ``text`` with ``repl``.

    Note:
        If your Python has a narrow unicode build ("USC-2"), only dingbats
        and miscellaneous symbols are replaced because Python isn't able
        to represent the unicode data for things like emoticons. Sorry!
    """
    return resources.RE_EMOJI.sub(repl, text)


def hashtags(text: str, repl: str = "_TAG_") -> str:
    """Replace all hashtags in ``text`` with ``repl``."""
    return resources.RE_HASHTAG.sub(repl, text)


def numbers(text: str, repl: str = "_NUMBER_") -> str:
    """Replace all numbers in ``text`` with ``repl``."""
    return resources.RE_NUMBER.sub(repl, text)


def phone_numbers(text: str, repl: str = "_PHONE_") -> str:
    """Replace all phone numbers in ``text`` with ``repl``."""
    return resources.RE_PHONE_NUMBER.sub(repl, text)


def urls(text: str, repl: str = "_URL_") -> str:
    """Replace all URLs in ``text`` with ``repl``."""
    return resources.RE_SHORT_URL.sub(repl, resources.RE_URL.sub(repl, text))


def user_handles(text: str, repl: str = "_USER_") -> str:
    """Replace all (Twitter-style) user handles in ``text`` with ``repl``."""
    return resources.RE_USER_HANDLE.sub(repl, text)



import pytest

from textacy.lang_id import identify_lang


@pytest.mark.parametrize(
    "text, exp",
    [
        ("This is a short example sentence in English.", "en"),
        ("Esta es una breve oración de ejemplo en español.", "es"),
        ("Ceci est un exemple court phrase en français.", "fr"),
        ("1", "un"),
        (" ", "un"),
        ("", "un"),
    ]
)
def test_lang_id(text, exp):
    obs = identify_lang(text)
    assert obs == exp



import pytest

from textacy.constants import RE_ACRONYM


@pytest.mark.parametrize(
    "item",
    [
        "LGTM",
        "U.S.A.",
        "PEP8",
        "LGBTQQI2S",
        "TF-IDF",
        "D3",
        "3D",
        "3-D",
        "3D-TV",
        "D&D",
        "PrEP",
        "H2SO4",
        "I/O",
        "WASPs",
        "G-8",
        "A-TReC",
    ]
)
def test_good_acronym_regex(item):
    assert item == RE_ACRONYM.search(item).group()


@pytest.mark.parametrize(
    "item",
    ["A", "GHz", "1a", "D o E", "Ms", "Ph.D", "3-Dim."]
)
def test_bad_acronym_regex(item):
    assert RE_ACRONYM.search(item) is None



"""
Collection of regular expressions and other (small, generally useful) constants.
"""
import pathlib
import re
from typing import Dict, Pattern, Set


DEFAULT_DATA_DIR: pathlib.Path = pathlib.Path(__file__).parent.resolve() / "data"

NUMERIC_ENT_TYPES: Set[str] = {
    "ORDINAL",
    "CARDINAL",
    "MONEY",
    "QUANTITY",
    "PERCENT",
    "TIME",
    "DATE",
}
SUBJ_DEPS: Set[str] = {"agent", "csubj", "csubjpass", "expl", "nsubj", "nsubjpass"}
OBJ_DEPS: Set[str] = {"attr", "dobj", "dative", "oprd"}
AUX_DEPS: Set[str] = {"aux", "auxpass", "neg"}

REPORTING_VERBS: Dict[str, Set[str]] = {
    "en": {
        "according",
        "accuse",
        "acknowledge",
        "add",
        "admit",
        "agree",
        "allege",
        "announce",
        "argue",
        "ask",
        "assert",
        "believe",
        "blame",
        "charge",
        "cite",
        "claim",
        "complain",
        "concede",
        "conclude",
        "confirm",
        "contend",
        "criticize",
        "declare",
        "decline",
        "deny",
        "describe",
        "disagree",
        "disclose",
        "estimate",
        "explain",
        "fear",
        "hope",
        "insist",
        "maintain",
        "mention",
        "note",
        "observe",
        "order",
        "predict",
        "promise",
        "recall",
        "recommend",
        "reply",
        "report",
        "say",
        "state",
        "stress",
        "suggest",
        "tell",
        "testify",
        "think",
        "urge",
        "warn",
        "worry",
        "write",
    },
    "es": {
        "afirmar",
        "anunciar",
        "añadir",
        "asegurar",
        "comentar",
        "confesar",
        "contestar",
        "decir",
        "escribir",
        "exclamar",
        "explicar",
        "gritar",
        "preguntar",
        "prometer",
        "quejar",
        "recordar",
    },
    "fr": {
        "affirmer",
        "ajouter",
        "annoncer",
        "conclure",
        "crier",
        "déclarer",
        "demander",
        "dire",
        "écrire",
        "exclamer",
        "expliquer",
        "insister",
        "lancer",
        "mentionner",
        "prétendre",
        "proclamer",
        "prononcer",
        "proposer",
        "remarquer",
        "répliquer",
        "répondre",
        "rétorquer",
        "soutenir",
        "suggérer",
    },
}

MATCHER_VALID_OPS: Set[str] = {"!", "+", "?", "*"}

RE_MATCHER_TOKPAT_DELIM: Pattern = re.compile(r"\s+")
RE_MATCHER_SPECIAL_VAL: Pattern = re.compile(r"^(int|bool)\([^: ]+\)$", flags=re.UNICODE)

RE_ACRONYM: Pattern = re.compile(
    r"(?:^|(?<=\W))"
    r"(?:"
    r"(?:(?:(?:[A-Z]\.?)+[a-z0-9&/-]?)+(?:[A-Z][s.]?|\ds?))"
    r"|"
    r"(?:\d(?:\-?[A-Z])+)"
    r")"
    r"(?:$|(?=\W))",
    flags=re.UNICODE,
)

RE_LINEBREAK: Pattern = re.compile(r"(\r\n|[\n\v])+")
RE_NONBREAKING_SPACE: Pattern = re.compile(r"[^\S\n\v]+", flags=re.UNICODE)

# regexes for cleaning up crufty terms
RE_DANGLING_PARENS_TERM: Pattern = re.compile(
    r"(?:\s|^)(\()\s{1,2}(.*?)\s{1,2}(\))(?:\s|$)", flags=re.UNICODE
)
RE_LEAD_TAIL_CRUFT_TERM: Pattern = re.compile(r"^[^\w(-]+|[^\w).!?]+$", flags=re.UNICODE)
RE_LEAD_HYPHEN_TERM: Pattern = re.compile(r"^-([^\W\d_])", flags=re.UNICODE)
RE_NEG_DIGIT_TERM: Pattern = re.compile(r"(-) (\d)", flags=re.UNICODE)
RE_WEIRD_HYPHEN_SPACE_TERM: Pattern = re.compile(
    r"(?<=[^\W\d]) (-[^\W\d])", flags=re.UNICODE
)
RE_WEIRD_APOSTR_SPACE_TERM: Pattern = re.compile(
    r"([^\W\d]+) ('[a-z]{1,2}\b)", flags=re.UNICODE
)

RE_ALNUM: Pattern = re.compile(r"[^\W_]+")



import os

import pytest

import textacy
from textacy.datasets.udhr import UDHR

DATASET = UDHR()


def _skipif():
    try:
        DATASET._check_data()
        return False
    except OSError:
        return True


pytestmark = pytest.mark.skipif(
    _skipif(),
    reason="UDHR dataset must be downloaded before running tests",
)


@pytest.mark.skip("No need to download a new dataset every time")
def test_download(tmpdir):
    dataset = UDHR(data_dir=str(tmpdir))
    dataset.download()
    assert os.path.isfile(dataset._index_filepath)
    assert os.path.isdir(dataset._texts_dirpath)


def test_oserror(tmpdir):
    dataset = UDHR(data_dir=str(tmpdir))
    with pytest.raises(OSError):
        _ = list(dataset.texts())


def test_texts():
    texts = list(DATASET.texts(limit=3))
    assert len(texts) > 0
    for text in texts:
        assert isinstance(text, str)


def test_texts_limit():
    for limit in (1, 5, 10):
        assert sum(1 for _ in DATASET.texts(limit=limit)) == limit


def test_records():
    for text, meta in DATASET.records(limit=3):
        assert isinstance(text, str)
        assert isinstance(meta, dict)


def test_records_lang():
    langs = ({"en"}, {"en", "es"})
    for lang in langs:
        records = list(DATASET.records(lang=lang, limit=10))
        assert all(meta["lang"] in lang for _, meta in records)


def test_bad_filters():
    bad_filters = (
        {"lang": "xx"},
        {"lang": ["en", "un"]},
    )
    for bad_filter in bad_filters:
        with pytest.raises(ValueError):
            list(DATASET.texts(**bad_filter))
    bad_filters = (
        {"lang": True},
        {"lang": textacy.load_spacy_lang("en_core_web_sm")},
    )
    for bad_filter in bad_filters:
        with pytest.raises(TypeError):
            list(DATASET.texts(**bad_filter))



from __future__ import annotations

import logging
import operator
import pathlib
import random
import re
from typing import Dict, Iterable, List, Optional, Tuple, Set

from cytoolz import itertoolz

import textacy
from textacy import io as tio

LOGGER = logging.getLogger(__name__)


class IsoLangResource:
    """
    Dataset based on the official ISO-639 code table,
    mapping all language code variations (639-1, 639-2, 639-3) to each other.

    Source: https://iso639-3.sil.org/code_tables/639/data
    """

    download_url = "https://iso639-3.sil.org/sites/iso639-3/files/downloads/iso-639-3.tab"
    filename = "iso-639-3.tsv"

    def __init__(self, data_dir: str | pathlib.Path):
        self.data_dir = textacy.utils.to_path(data_dir).resolve()

    def download(self, force: bool = False):
        """
        Args:
            force: If True, always download a new copy of the dataset; otherwise,
                only download dataset if it doesn't already exist on disk.
        """
        tio.download_file(
            self.download_url,
            filename=self.filename,
            dirpath=self.data_dir,
            force=force,
        )

    def load(self, exclude: Optional[Set[str]] = None) -> Dict[str, str]:
        """
        Args:
            exclude

        Returns:
            Dict[str, str]
        """
        rows = tio.read_csv(
            self.data_dir.joinpath(self.filename),
            delimiter="\t",
            fieldnames=[
                "Id",
                "Part2B",
                "Part2T",
                "Part1",
                "Scope",
                "Language_Type",
                "Ref_Name",
                "Comment",
            ],
            quoting=1,
        )
        lang_map = {
            row["Id"]: row["Part1"]
            for row in rows
            if row.get("Part1") and (exclude is None or row["Part1"] not in exclude)
        }
        LOGGER.info(
            "loaded IsoLangResource data:\n%s ...",
            sorted(lang_map.items())[:5],
        )
        return lang_map


class DSLCCDataset:
    """
    Dataset based on two multilingual collections of short excerpts of journalistic texts,
    focused on language groups that are very similar and thus more difficult
    to correctly identify.

    Source: http://ttg.uni-saarland.de/resources/DSLCC
    """

    def __init__(self, data_dir: str | pathlib.Path):
        self.data_dir = textacy.utils.to_path(data_dir).resolve()

    def download(self, force: bool = False):
        """
        Args:
            force: If True, always download a new copy of the dataset; otherwise,
                only download dataset if it doesn't already exist on disk.
        """
        for version in [3, 4]:
            name = f"dslcc{version}"
            url = f"http://scholar.harvard.edu/files/malmasi/files/{name}.zip"
            fpath = tio.download_file(url, dirpath=self.data_dir, force=force)
            if fpath:
                tio.unpack_archive(fpath, extract_dir=self.data_dir.joinpath(name))

    def load(self, langs: Set[str], min_len: int = 25) -> List[Tuple[str, str]]:
        """
        Args:
            langs
            min_len: Minimum text length in *chars* for a given example to be included.

        Returns:
            Sequence of (text, lang) examples.
        """
        data = []
        fstubs = [
            "dslcc3/train/task1-train.txt",
            "dslcc3/train/task1-dev.txt",
            "dslcc4/DSL-TRAIN.txt",
            "dslcc4/DSL-DEV.txt",
        ]
        for fstub in fstubs:
            filepath = self.data_dir.joinpath(fstub)
            lines = tio.read_text(filepath, mode="rt", encoding="utf-8", lines=True)
            for line in lines:
                if not line.strip():
                    continue
                try:
                    text, lang = line.split("\t")
                    if (
                        lang[:2] in langs
                        and itertoolz.count(c for c in text if c.isalnum()) >= min_len
                    ):
                        data.append((text, lang[:2]))
                except Exception:
                    LOGGER.debug("bad line in data")
                    pass
        data = sorted(set(data), key=operator.itemgetter(1))
        LOGGER.info("loaded DSLCCDataset data:\n%s ...", data[:3])
        return data


class TatoebaDataset:

    download_url = "http://downloads.tatoeba.org/exports/sentences.tar.bz2"

    def __init__(self, data_dir: str | pathlib.Path):
        self.data_dir = textacy.utils.to_path(data_dir).resolve()

    def download(self, force: bool = False):
        """
        Args:
            force: If True, always download a new copy of the dataset; otherwise,
                only download dataset if it doesn't already exist on disk.

        Note:
            Source: https://tatoeba.org/eng/downloads
        """
        fpath = tio.download_file(self.download_url, dirpath=self.data_dir, force=force)
        if fpath:
            tio.unpack_archive(fpath, extract_dir=self.data_dir)

    def load(
        self,
        iso_lang_map: Dict[str, str],
        min_len: int = 25,
    ) -> List[Tuple[str, str]]:
        """
        Args:
            iso_lang_map
            min_len: Minimum text length in *chars* for a given example to be included.

        Returns:
            Sequence of (text, lang) examples.
        """
        rows = tio.read_csv(
            self.data_dir.joinpath("sentences.csv"),
            fieldnames=["sent_id", "iso-639-3", "text"],
            delimiter="\t",
            quoting=1,
        )
        data = [
            (row["text"], iso_lang_map[row["iso-639-3"]])
            for row in rows
            if row["iso-639-3"] in iso_lang_map
            and itertoolz.count(char for char in row["text"] if char.isalnum()) >= min_len
        ]
        LOGGER.info("loaded TatoebaDataset data:\n%s ...", data[:3])
        return data


class Wili2018Dataset:
    """
    Dataset based on paragraphs from Wikipedia in 230+ languages.

    Source: https://zenodo.org/record/841984

    References:
        Thoma, Martin. "The WiLI benchmark dataset for written language identification."
        arXiv preprint arXiv:1801.07779 (2018).
    """

    download_url = "https://zenodo.org/record/841984/files/wili-2018.zip?download=1"

    def __init__(self, data_dir: str | pathlib.Path):
        self.data_dir = textacy.utils.to_path(data_dir).resolve()

    def download(self, force: bool = False):
        """
        Args:
            force: If True, always download a new copy of the dataset; otherwise,
                only download dataset if it doesn't already exist on disk.
        """
        fpath = tio.download_file(
            self.download_url, dirpath=self.data_dir, force=force
        )
        if fpath:
            tio.unpack_archive(fpath, extract_dir=self.data_dir)

    def load(
        self,
        iso_lang_map: Dict[str, str],
        min_len: int = 25,
    ) -> List[Tuple[str, str]]:
        """
        Args:
            iso_lang_map
            min_len: Minimum text length in *chars* for a given example to be included.

        Returns:
            Sequence of (text, lang) examples.
        """
        data = []
        # we'll combine train/test from individual datasets
        # and instead split on the full, aggregated dataset
        for subset in ("train", "test"):
            text_lines = tio.read_text(
                self.data_dir.joinpath(f"x_{subset}.txt"), lines=True
            )
            lang_lines = tio.read_text(
                self.data_dir.joinpath(f"y_{subset}.txt"), lines=True
            )
            texts = (line.strip() for line in text_lines)
            langs = (line.strip() for line in lang_lines)
            data.extend(
                (text, iso_lang_map[lang])
                for text, lang in zip(texts, langs)
                if lang in iso_lang_map
                and itertoolz.count(char for char in text if char.isalnum()) >= min_len
            )
        LOGGER.info("loaded Wili2018Dataset data:\n%s ...", data[:3])
        return data


class UDDataset:
    """
    Source: https://lindat.mff.cuni.cz/repository/xmlui/handle/11234/1-3424

    References:
        Zeman, Daniel; Nivre, Joakim; Abrams, Mitchell; et al., 2020, Universal Dependencies 2.7,
        LINDAT/CLARIAH-CZ digital library at the Institute of Formal and Applied Linguistics (ÚFAL),
        Faculty of Mathematics and Physics, Charles University,
        http://hdl.handle.net/11234/1-3424.
    """

    download_url = "https://lindat.mff.cuni.cz/repository/xmlui/bitstream/handle/11234/1-3424/ud-treebanks-v2.7.tgz"

    def __init__(self, data_dir: str | pathlib.Path):
        self.data_dir = textacy.utils.to_path(data_dir).resolve()

    def download(self, force: bool = False):
        """
        Args:
            force: If True, always download a new copy of the dataset; otherwise,
                only download dataset if it doesn't already exist on disk.
        """
        fpath = tio.download_file(
            self.download_url, dirpath=self.data_dir, force=force
        )
        if fpath:
            tio.unpack_archive(fpath, extract_dir=self.data_dir)

    def load(self, langs: Set[str], min_len: int = 25) -> List[Tuple[str, str]]:
        """
        Args:
            langs
            min_len: Minimum text length in *chars* for a given example to be included.

        Returns:
            Sequence of (text, lang) examples.
        """
        data = []
        match_regex = r"ud-(train|test|dev)\.txt"
        for fpath in tio.get_filepaths(
            self.data_dir, match_regex=match_regex, recursive=True
        ):
            fname = pathlib.Path(fpath).name
            lang, _ = fname.split("_", maxsplit=1)
            if lang not in langs:
                continue

            with open(fpath, mode="rt") as f:
                text = f.read()
            if "\n" in text:
                data.extend(
                    (text_segment, lang)
                    for text_segment in re.split(r"\n+", text)
                    if len(text_segment) >= min_len
                )
            else:
                data.extend(
                    (text_segment, lang)
                    for text_segment in _randomly_segment_text(text, (50, 1000))
                    if len(text_segment) >= min_len
                )
        LOGGER.info("loaded TatoebaDataset data:\n%s ...", data[:3])
        return data


def _randomly_segment_text(text: str, len_range: Tuple[int, int]) -> Iterable[str]:
    min_len, max_len = len_range
    idxs = []
    idx = 0
    while idx < len(text):
        idxs.append(idx)
        idx += random.randint(min_len, max_len)
    idxs.append(len(text))
    for idx_start, idx_end in itertoolz.sliding_window(2, idxs):
        yield text[idx_start : idx_end]



import numpy as np
import pytest
from sklearn.decomposition import NMF, LatentDirichletAllocation, TruncatedSVD

from textacy import Corpus
from textacy.tm import TopicModel
from textacy.representations import Vectorizer


@pytest.fixture(scope="module")
def term_lists():
    texts = [
        "Mary had a little lamb. Its fleece was white as snow.",
        "Everywhere that Mary went the lamb was sure to go.",
        "It followed her to school one day, which was against the rule.",
        "It made the children laugh and play to see a lamb at school.",
        "And so the teacher turned it out, but still it lingered near.",
        "It waited patiently about until Mary did appear.",
        "Why does the lamb love Mary so? The eager children cry.",
        "Mary loves the lamb, you know, the teacher did reply.",
    ]
    corpus = Corpus("en_core_web_sm", data=texts)
    term_lists_ = [
        (term.text.lower() for term in doc._.extract_terms(ngs=1, ents=None))
        for doc in corpus
    ]
    return term_lists_


@pytest.fixture(scope="module")
def vectorizer():
    vectorizer_ = Vectorizer(
        tf_type="linear",
        idf_type="smooth",
        norm=None,
        min_df=1,
        max_df=1.0,
        max_n_terms=None,
    )
    return vectorizer_


@pytest.fixture(scope="module")
def doc_term_matrix(term_lists, vectorizer):
    doc_term_matrix_ = vectorizer.fit_transform(term_lists)
    return doc_term_matrix_


@pytest.fixture(scope="module")
def model(doc_term_matrix):
    model_ = TopicModel("nmf", n_topics=5)
    model_.fit(doc_term_matrix)
    return model_


def test_n_topics():
    for model in ["nmf", "lda", "lsa"]:
        assert TopicModel(model, n_topics=20).n_topics == 20


def test_duck_typing():
    class TrainedDummyModel():
        def __init__(self):
            self.n_topics = 5
            self.components_ = np.array([[0,0,0,1], [1,0,0,0]])

        def transform(self, text):
            return text
    dummy = TrainedDummyModel()
    tmodel = TopicModel(dummy)

    assert tmodel.n_topics == dummy.n_topics
    assert tmodel.model.transform == dummy.transform
    np.testing.assert_array_equal(tmodel.model.components_, dummy.components_)


def test_init_model():
    expecteds = (NMF, LatentDirichletAllocation, TruncatedSVD)
    models = ["nmf", "lda", "lsa"]
    for model, expected in zip(models, expecteds):
        assert isinstance(TopicModel(model).model, expected)


def test_save_load(tmpdir, model):
    filepath = str(tmpdir.join("model.pkl"))
    expected = model.model.components_
    model.save(filepath)
    tmp_model = TopicModel.load(filepath)
    observed = tmp_model.model.components_
    assert observed.shape == expected.shape
    assert np.equal(observed, expected).all()


def test_transform(doc_term_matrix, model):
    expected = (doc_term_matrix.shape[0], model.n_topics)
    observed = model.transform(doc_term_matrix).shape
    assert observed == expected


@pytest.mark.skip(reason="this sometimes fails randomly, reason unclear...")
def test_get_doc_topic_matrix(doc_term_matrix, model):
    expected = np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0])
    observed = model.get_doc_topic_matrix(doc_term_matrix, normalize=True).sum(axis=1)
    assert np.equal(observed, expected).all()


def test_get_doc_topic_matrix_nonnormalized(doc_term_matrix, model):
    expected = model.transform(doc_term_matrix)
    observed = model.get_doc_topic_matrix(doc_term_matrix, normalize=False)
    assert np.equal(observed, expected).all()


def test_top_topic_terms_topics(vectorizer, model):
    assert (
        len(list(model.top_topic_terms(vectorizer.id_to_term, topics=-1)))
        == model.n_topics
    )
    assert len(list(model.top_topic_terms(vectorizer.id_to_term, topics=0))) == 1
    observed = [
        topic_idx
        for topic_idx, _ in model.top_topic_terms(
            vectorizer.id_to_term, topics=(1, 2, 3)
        )
    ]
    expected = [1, 2, 3]
    assert observed == expected


def test_top_topic_terms_top_n(vectorizer, model):
    assert (
        len(
            list(model.top_topic_terms(vectorizer.id_to_term, topics=0, top_n=10))[0][1]
        )
        == 10
    )
    assert (
        len(list(model.top_topic_terms(vectorizer.id_to_term, topics=0, top_n=5))[0][1])
        == 5
    )


def test_top_topic_terms_weights(vectorizer, model):
    observed = list(
        model.top_topic_terms(vectorizer.id_to_term, topics=-1, top_n=10, weights=True)
    )
    assert isinstance(observed[0][1][0], tuple)
    for topic_idx, term_weights in observed:
        for i in range(len(term_weights) - 1):
            assert term_weights[i][1] >= term_weights[i + 1][1]


def _xfailif():
    try:
        import matplotlib.pyplot as plt
        return False
    except ImportError:
        return True


@pytest.mark.xfail(
    _xfailif(),
    reason="matplotlib is an optional dependency, but required for this viz")
def test_termite_plot(model, vectorizer, doc_term_matrix):
    model.termite_plot(
        doc_term_matrix, vectorizer.id_to_term,
        topics=-1, n_terms=25, sort_terms_by="seriation")



import datetime
import os

import pytest

from textacy.datasets.reddit_comments import RedditComments

DATASET = RedditComments()

pytestmark = pytest.mark.skipif(
    not DATASET.filepaths,
    reason="RedditComments dataset must be downloaded before running tests",
)


@pytest.mark.skip("No need to download a new dataset every time")
def test_download(tmpdir):
    dataset = RedditComments(data_dir=str(tmpdir))
    dataset.download()
    assert all(os.path.isfile(filepath) for filepath in dataset.filepaths)


def test_oserror(tmpdir):
    dataset = RedditComments(data_dir=str(tmpdir))
    with pytest.raises(OSError):
        _ = list(dataset.texts())


def test_texts():
    texts = list(DATASET.texts(limit=3))
    assert len(texts) > 0
    for text in texts:
        assert isinstance(text, str)


def test_texts_limit():
    for limit in (1, 5, 10):
        assert sum(1 for _ in DATASET.texts(limit=limit)) == limit


def test_texts_min_len():
    for min_len in (100, 200, 500):
        assert all(
            len(text) >= min_len for text in DATASET.texts(min_len=min_len, limit=10)
        )


def test_records():
    for text, meta in DATASET.records(limit=3):
        assert isinstance(text, str)
        assert isinstance(meta, dict)


def test_records_subreddit():
    subreddits = ({"politics"}, {"politics", "programming"})
    for subreddit in subreddits:
        records = list(DATASET.records(subreddit=subreddit, limit=10))
        assert all(
            meta["subreddit"] in subreddit
            for text, meta in records
        )


def test_records_date_range():
    date_ranges = (("2007-10", "2007-12"), ["2007-10-01", "2008-01-01"])
    for date_range in date_ranges:
        records = list(DATASET.records(date_range=date_range, limit=10))
        assert all(
            date_range[0] <= meta["created_utc"] < date_range[1]
            for text, meta in records
        )


def test_records_score_range():
    score_ranges = ([-10, 10], (5, 100))
    for score_range in score_ranges:
        records = list(DATASET.records(score_range=score_range, limit=10))
        assert all(
            score_range[0] <= meta["score"] < score_range[1]
            for text, meta in records
        )


def test_bad_filters():
    bad_filters = (
        {"min_len": -1},
    )
    for bad_filter in bad_filters:
        with pytest.raises(ValueError):
            list(DATASET.texts(**bad_filter))
    bad_filters = (
        {"date_range": "2016-01-01"},
        {"date_range": (datetime.date(2007, 10, 1), datetime.date(2007, 11, 1))},
        {"score_range": 10},
        {"score_range": ["low", "high"]},
    )
    for bad_filter in bad_filters:
        with pytest.raises(TypeError):
            list(DATASET.texts(**bad_filter))



import pytest

from spacy.tokens import Token

from textacy import extract
from textacy import load_spacy_lang


@pytest.fixture(scope="module")
def spacy_lang():
    return load_spacy_lang("en_core_web_sm")


@pytest.mark.parametrize(
    "token",
    [
        "LGTM",
        "U.S.A.",
        "PEP8",
        "LGBTQQI2S",
        "TF-IDF",
        "D3",
        "3D",
        "3-D",
        "3D-TV",
        "D&D",
        "PrEP",
        "H2SO4",
        "I/O",
        "WASPs",
        "G-8",
        "A-TReC",
    ]
)
def test_is_acronym_good(token):
    assert extract.acros.is_acronym(token)


@pytest.mark.parametrize(
    "token",
    [
        "A",
        "GHz",
        "1a",
        "D o E",
        "Ms",
        "Ph.D",
        "3-Dim.",
        "the",
        "FooBar",
        "1",
        " ",
        "",
    ]
)
def test_is_acronym_bad(token):
    assert not extract.acros.is_acronym(token)


@pytest.mark.parametrize(
    "token,exclude,expected",
    [
        ("NASA", {"NASA"}, False),
        ("NASA", {"CSA", "ISS"}, True),
        ("NASA", None, True)
    ]
)
def test_is_acronym_exclude(token, exclude, expected):
    assert extract.acros.is_acronym(token, exclude=exclude) == expected


@pytest.mark.parametrize(
    "text, exp",
    [
        ("I want to work for NASA when I grow up, but not NOAA.", ["NASA", "NOAA"]),
        ("I want to live in the U.S. Do you?", ["U.S."]),
    ]
)
def test_acronyms(spacy_lang, text, exp):
    doc = spacy_lang(text)
    obs = list(extract.acronyms(doc))
    assert all(isinstance(tok, Token) for tok in obs)
    assert [tok.text for tok in obs] == exp


class TestAcronymsAndDefinitions:

    @pytest.mark.parametrize(
        "text, exp",
        [
            (
                "I want to work for NASA (National Aeronautics and Space Administration) when I grow up.",
                {"NASA": "National Aeronautics and Space Administration"},
            ),
            (
                "I want to work for NASA — National Aeronautics and Space Administration — when I grow up.",
                {"NASA": "National Aeronautics and Space Administration"},
            ),
            (
                "I want to work for NASA -- National Aeronautics and Space Administration -- when I grow up.",
                {"NASA": "National Aeronautics and Space Administration"},
            ),
            (
                "I want to work for NASA when I grow up.",
                {"NASA": ""},
            ),
            (
                "I want to live in the U.S.A. (United States of America) when I grow up.",
                # TODO: fix whatever error is failing to find the definition here
                # {"U.S.A.": "United States of America"},
                {"U.S.A.": ""},
            ),
            (
                "I want to be a W.A.S.P. (White Anglo Saxon Protestant) when I grow up.",
                # TODO: fix whatever error is failing to find the definition here
                # {"W.A.S.P": "White Anglo Saxon Protestant"},
                {"W.A.S.P.": ""},
            ),
        ]
    )
    def test_default(self, spacy_lang, text, exp):
        obs = extract.acronyms_and_definitions(spacy_lang(text))
        assert obs == exp

    @pytest.mark.parametrize(
        "text, known, exp",
        [
            (
                "I want to work for NASA when I grow up.",
                {"NASA": "National Aeronautics and Space Administration"},
                {"NASA": "National Aeronautics and Space Administration"},
            ),
            (
                "I want to work for NASA when I grow up.",
                {"NASA": "National Auto Sport Association"},
                {"NASA": "National Auto Sport Association"},
            ),
            (
                "I want to work for NASA (National Aeronautics and Space Administration) when I grow up.",
                {"NASA": "National Auto Sport Association"},
                {"NASA": "National Auto Sport Association"},
            ),
        ],
    )
    def test_default(self, spacy_lang, text, known, exp):
        obs = extract.acronyms_and_definitions(spacy_lang(text), known_acro_defs=known)
        assert obs == exp



import datetime
import os

import pytest

from textacy.datasets.oxford_text_archive import OxfordTextArchive

DATASET = OxfordTextArchive()

pytestmark = pytest.mark.skipif(
    DATASET.metadata is None,
    reason="OxfordTextArchive dataset must be downloaded before running tests",
)


@pytest.mark.skip("No need to download a new dataset every time")
def test_download(tmpdir):
    dataset = OxfordTextArchive(data_dir=str(tmpdir))
    dataset.download()
    assert os.path.isfile(dataset._metadata_filepath)
    assert os.path.isdir(dataset._text_dirpath)


def test_oserror(tmpdir):
    dataset = OxfordTextArchive(data_dir=str(tmpdir))
    with pytest.raises(OSError):
        _ = list(dataset.texts())


def test_metadata():
    assert DATASET.metadata is not None
    for record in DATASET.records(limit=10):
        pass
    assert all(entry.get("id") for entry in DATASET.metadata.values())
    assert not any(entry.get("text") for entry in DATASET.metadata.values())


def test_texts():
    for text in DATASET.texts(limit=3):
        assert isinstance(text, str)


def test_texts_limit():
    for limit in (1, 5, 10):
        assert sum(1 for _ in DATASET.texts(limit=limit)) == limit


def test_texts_min_len():
    for min_len in (100, 200, 1000):
        assert all(
            len(text) >= min_len for text in DATASET.texts(min_len=min_len, limit=10)
        )


def test_records():
    for text, meta in DATASET.records(limit=3):
        assert isinstance(text, str)
        assert isinstance(meta, dict)


def test_records_author():
    authors = ({"London, Jack"}, {"Burroughs, Edgar Rice", "Wells, H.G. (Herbert George)"})
    for author in authors:
        records = list(DATASET.records(author=author, limit=3))
        assert len(records) >= 1
        assert any(
            athr in author
            for text, meta in records
            for athr in meta["author"]
        )


def test_records_date_range():
    date_ranges = (["1840-01-01", "1860-01-01"], ("1500-01-01", "1600-01-01"))
    for date_range in date_ranges:
        records = list(DATASET.records(date_range=date_range, limit=3))
        assert len(records) >= 1
        assert all(
            date_range[0] <= meta["year"] < date_range[1]
            for text, meta in records
        )


def test_bad_filters():
    bad_filters = (
        {"author": "Burton DeWilde"},
        {"min_len": -1},
    )
    for bad_filter in bad_filters:
        with pytest.raises(ValueError):
            list(DATASET.texts(**bad_filter))
    bad_filters = (
        {"date_range": "2016-01-01"},
        {"date_range": (datetime.date(1800, 1, 1), datetime.date(1900, 1, 1))},
    )
    for bad_filter in bad_filters:
        with pytest.raises(TypeError):
            list(DATASET.texts(**bad_filter))



from __future__ import annotations

from typing import Iterable, List, Sequence, Tuple

import sklearn.preprocessing
import thinc
from tqdm import tqdm
from thinc.api import Model

from . import models


def train_model(
    model: Model,
    *,
    train: Sequence[Tuple[str, str]],
    test: Sequence[Tuple[str, str]],
    n_iter: int,
    batch_size: int | thinc.types.Generator = 32,
    learn_rate: float | List[float] | thinc.types.Generator = 0.001,
) -> Model:
    """
    Args:
        model
        train
        test
        n_iter
        batch_size
        learn_rate
    """
    # binarize language labels
    # NOTE: thinc seems to require type "float32" arrays for training labels
    # errors otherwise... :/
    lb = sklearn.preprocessing.LabelBinarizer()
    lb.fit([lang for _, lang in train])
    # THIS NEXT LINE IS CRITICAL: we need to save the training class labels
    # but don't want to keep this label binarizer around; so, add it to the model
    model.layers[-1].attrs["classes"] = lb.classes_

    Y_train = lb.transform([lang for _, lang in train]).astype("float32")
    Y_test = lb.transform([lang for _, lang in test])

    # make sure data is on the right device?
    # Y_train = self.model.ops.asarray(Y_train)
    # Y_test = self.model.ops.asarray(Y_test)

    X_train = [text for text, _ in train]
    X_test = [text for text, _ in test]

    losser = thinc.api.CategoricalCrossentropy(normalize=True)
    optimizer = thinc.api.Adam(learn_rate)

    model.initialize(X=X_train[:10], Y=Y_train[:10])
    print(f"{'epoch':>5}  {'loss':>8}  {'score':>8}")
    # iterate over epochs
    for n in range(n_iter):
        loss = 0.0
        # iterate over batches
        batches = model.ops.multibatch(batch_size, X_train, Y_train, shuffle=True)
        for X, Y in tqdm(batches, leave=False):
            Yh, backprop = model.begin_update(X)
            dYh, loss_batch = losser(Yh, Y)
            loss += loss_batch
            backprop(dYh)
            model.finish_update(optimizer)
            optimizer.step_schedules()

        if optimizer.averages:
            with model.use_params(optimizer.averages):
                score = evaluate_model(model, X_test=X_test, Y_test=Y_test, batch_size=1000)
        else:
            score = evaluate_model(model, X_test=X_test, Y_test=Y_test, batch_size=1000)
        print(f"{n:>5}  {loss:>8.3f}  {score:>8.3f}")

    if optimizer.averages:
        with model.use_params(optimizer.averages):
            pred_langs = models.get_model_preds(
                model, X_test, model.layers[-1].attrs["classes"]
            )
    else:
        pred_langs = models.get_model_preds(
            model, X_test, model.layers[-1].attrs["classes"]
        )
    true_langs = list(lb.inverse_transform(Y_test))
    print(sklearn.metrics.classification_report(true_langs, pred_langs))
    return model


def evaluate_model(
    model,
    *,
    X_test: Sequence[str],
    Y_test : thinc.types.Array2d,
    batch_size: int,
) -> float:
    """
    Args:
        model
        X_test
        Y_test
        batch_size
    """
    correct = 0
    total = 0
    for X, Y in model.ops.multibatch(batch_size, X_test, Y_test):
        Yh = model.predict(X)
        for yh, y in zip(Yh, Y):
            correct += (y.argmax(axis=0) == yh.argmax(axis=0)).sum()
        total += len(Y)
    return float(correct / total)


def decaying_cyclic_triangular(
    decaying: thinc.types.Generator,
    cyclic_triangular: thinc.types.Generator,
    *,
    min_lr: float = 0.0,
) -> Iterable[float]:
    """
    Custom "schedule" that multiplicatively combines two built-in thinc schedules:
    :func:`thinc.api.decaying()` and :func:`thinc.api.cyclic_triangular()`

    Args:
        decaying
        cyclic_triangular
        min_lr: Minimum learning rate, after combination. Combined values that are
            smaller than this will be set equal to it.
    """
    for lr_decay, lr_ct in zip(decaying, cyclic_triangular):
        yield max(lr_decay * lr_ct, min_lr)



import numpy as np
import pytest
import scipy.sparse as sp
from spacy.tokens import Doc, Span

import textacy.datasets
from textacy import Corpus
from textacy import extract, preprocessing
from textacy import load_spacy_lang, make_spacy_doc
from textacy.extract import keyterms as kt
from textacy.text_stats import TextStats
from textacy.tm import TopicModel
from textacy.representations import Vectorizer

DATASET = textacy.datasets.CapitolWords()

pytestmark = pytest.mark.skipif(
    DATASET.filepath is None,
    reason="CapitolWords dataset must be downloaded before running tests",
)


@pytest.fixture(scope="module")
def text():
    return next(DATASET.texts(speaker_name={"Bernie Sanders"}, limit=1)).strip()


@pytest.fixture(scope="module")
def doc(text):
    spacy_lang = load_spacy_lang("en_core_web_sm")
    return make_spacy_doc(text, lang=spacy_lang)


@pytest.fixture(scope="module")
def corpus():
    records = DATASET.records(speaker_name={"Bernie Sanders"}, limit=10)
    corpus = Corpus("en_core_web_sm", data=records)
    return corpus


def test_streaming_functionality(corpus):
    assert isinstance(DATASET, textacy.datasets.base.Dataset)
    assert isinstance(corpus, Corpus)


def test_vectorization_and_topic_modeling_functionality(corpus):
    n_topics = 10
    top_n = 10
    vectorizer = Vectorizer(
        tf_type="linear",
        idf_type="smooth",
        norm=None,
        min_df=2,
        max_df=0.95,
    )
    doc_term_matrix = vectorizer.fit_transform(
        (
            (term.text for term in doc._.extract_terms(ngs=1, ents=True, ncs=True))
            for doc in corpus
        )
    )
    model = TopicModel("nmf", n_topics=n_topics)
    model.fit(doc_term_matrix)
    doc_topic_matrix = model.transform(doc_term_matrix)
    assert isinstance(doc_term_matrix, sp.csr_matrix)
    assert isinstance(doc_topic_matrix, np.ndarray)
    assert doc_topic_matrix.shape[1] == n_topics
    for topic_idx, top_terms in model.top_topic_terms(
        vectorizer.id_to_term, top_n=top_n
    ):
        assert isinstance(topic_idx, int)
        assert len(top_terms) == top_n


def test_corpus_functionality(corpus):
    assert isinstance(corpus[0], Doc)
    assert list(
        corpus.get(lambda doc: doc._.meta["speaker_name"] == "Bernie Sanders")
    )


def test_plaintext_functionality(text):
    preprocessed_text = preprocessing.normalize.whitespace(text)
    preprocessed_text = preprocessing.remove.punctuation(text)
    preprocessed_text = preprocessed_text.lower()
    assert all(char.islower() for char in preprocessed_text if char.isalpha())
    assert all(char.isalnum() or char.isspace() for char in preprocessed_text)
    keyword = "America"
    kwics = extract.keyword_in_context(text, keyword, window_width=35)
    for pre, kw, post in kwics:
        assert kw == keyword
        assert isinstance(pre, str)
        assert isinstance(post, str)


def test_extract_functionality(doc):
    bigrams = list(
        extract.ngrams(doc, 2, filter_stops=True, filter_punct=True, filter_nums=False)
    )[:10]
    for bigram in bigrams:
        assert isinstance(bigram, Span)
        assert len(bigram) == 2

    trigrams = list(
        extract.ngrams(doc, 3, filter_stops=True, filter_punct=True, min_freq=2)
    )[:10]
    for trigram in trigrams:
        assert isinstance(trigram, Span)
        assert len(trigram) == 3

    nes = list(
        extract.entities(doc, drop_determiners=False, exclude_types="numeric")
    )[:10]
    for ne in nes:
        assert isinstance(ne, Span)
        assert ne.label_
        assert ne.label_ != "QUANTITY"

    regex_matches = list(extract.regex_matches(doc, "Mr\. Speaker"))[:10]
    for match in regex_matches:
        assert isinstance(match, Span)

    stmts = list(extract.semistructured_statements(doc, entity="I", cue="be"))[:10]
    for stmt in stmts:
        assert isinstance(stmt, list)
        assert isinstance(stmt[0], str)
        assert len(stmt) == 3

    kts = kt.textrank(doc, topn=10)
    for keyterm in kts:
        assert isinstance(keyterm, tuple)
        assert isinstance(keyterm[0], str)
        assert isinstance(keyterm[1], float)
        assert keyterm[1] > 0.0


def test_text_stats_functionality(doc):
    ts = TextStats(doc)

    for attr in ["n_words", "n_syllables", "n_chars"]:
        assert hasattr(ts, attr)
        assert isinstance(getattr(ts, attr), int)

    for attr in ["entropy", "flesch_kincaid_grade_level", "flesch_reading_ease", "lix"]:
        assert hasattr(ts, attr)
        assert isinstance(getattr(ts, attr), float)



import pytest

from textacy.similarity import hybrid


@pytest.mark.parametrize(
    "s1, s2, exp",
    [
        ("abcd", "abcd", 1.0),
        ("abdc", "wxyz", 0.0),
        ("abcd", "", 0.0),
        ("", "", 0.0),
        ("abcd", "ABCD", 1.0),
        ("abcd", "bacd", 0.5),
        ("abcd", "abXcd", 0.8),
        ("abcd", "abcdefgh", 0.5),
        ("abcdefgh", "abcd", 0.5),
        (list("abcd"), list("abcd"), 1.0),
        (list("abcd"), list("ABCD"), 1.0),
        ("abcd", "abcd efgh", 0.4444),
        (["abcd"], ["abcd", "efgh"], 0.4444),
        (["abcd", "wxyz"], ["abcd", "efgh"], 0.5555),
    ]
)
def test_token_sort_ratio(s1, s2, exp):
    obs = hybrid.token_sort_ratio(s1, s2)
    assert isinstance(obs, float)
    assert 0.0 <= obs <= 1.0
    assert obs == pytest.approx(exp, rel=0.01)


@pytest.mark.parametrize(
    "s1, s2, exp",
    [
        (["abcd", "efgh", "ijkl"], ["abcd", "efgh", "ijkl"], 1.0),
        (["abcd", "efgh", "ijkl"], ["opqr", "stuv", "wxyz"], 0.0),
        (["abcd", "efgh", "ijkl"], [], 0.0),
        ([], [], 0.0),
        (["abcd", "efgh", "ijkl"], ["Abcd", "efgh", "ijkl"], 0.9166),
        (["abcd", "efgh", "ijkl"], ["Abcd", "Efgh", "Ijkl"], 0.75),
        (["abcd", "efgh", "ijkl"], ["abcd", "efgh"], 0.8333),
        (["abcd", "efgh"], ["abcd", "efgh", "ijkl"], 0.8333),
        (["abcd", "efgh", "ijkl"], ["ABCD", "EFGH", "IJKL"], 0.0),
    ]
)
def test_monge_elkan(s1, s2, exp):
    obs = hybrid.monge_elkan(s1, s2)
    assert isinstance(obs, float)
    assert 0.0 <= obs <= 1.0
    assert obs == pytest.approx(exp, rel=0.01)



"""
:mod:`textacy.types`: Definitions for common object types used throughout the package.
"""
import collections
from pathlib import Path
from typing import Callable, Iterable, Union

from spacy.language import Language
from spacy.tokens import Doc, Span, Token


# Record = Tuple[str, dict]  => let's use a namedtuple instead
Record = collections.namedtuple("Record", ["text", "meta"])
DocData = Union[str, Record, Doc]
CorpusData = Union[str, Record, Doc, Iterable[str], Iterable[Record], Iterable[Doc]]

LangLike = Union[str, Path, Language]
LangLikeInContext = Union[
    str,
    Path,
    Language,
    Callable[[str], str],
    Callable[[str], Path],
    Callable[[str], Language],
]

DocLike = Union[Doc, Span]
SpanLike = Union[Span, Token]

PathLike = Union[str, Path]

DocLikeToSpans = Callable[[DocLike], Iterable[Span]]



import pytest

import textacy
import textacy.resources


RESOURCE = textacy.resources.ConceptNet()

pytestmark = pytest.mark.skipif(
    RESOURCE.filepath is None,
    reason="ConceptNet resource must be downloaded before running tests",
)


@pytest.fixture(scope="module")
def spacy_doc():
    text = "The quick brown fox jumps over the lazy dog."
    return textacy.make_spacy_doc(text, lang="en_core_web_sm")


@pytest.mark.skip("No need to download a new resource every time")
def test_download(tmpdir):
    resource = textacy.resources.ConceptNet(data_dir=str(tmpdir))
    resource.download()
    assert resource._filepath.is_file()


def test_oserror(tmpdir):
    resource = textacy.resources.ConceptNet(data_dir=str(tmpdir))
    with pytest.raises(OSError):
        _ = resource.antonyms


class TestGetRelationValues:

    def test_empty_results(self):
        invalid_senses = ["x", "y", "PROPN", "PUNCT"]
        for sense in invalid_senses:
            assert RESOURCE.get_antonyms("love", lang="en", sense=sense) == []
        missing_terms = ["foobarbatbaz", "burton dewilde", "natural language processing"]
        for term in missing_terms:
            assert RESOURCE.get_antonyms(term, lang="en", sense="n") == []

    def test_value_errors(self):
        bad_tls = [
            ("love", None, None),
            ("love", "en", None),
            ("love", None, "n"),
            ("love", "un", "n"),
        ]
        for term, lang, sense in bad_tls:
            with pytest.raises(ValueError):
                RESOURCE._get_relation_values(
                    RESOURCE.antonyms, term, lang=lang, sense=sense)

    def test_type_errors(self, spacy_doc):
        bad_tls = [
            (1, "en", "n"),
            (None, "en", "v"),
            (spacy_doc, "en", "a"),
        ]
        for term, lang, sense in bad_tls:
            with pytest.raises(TypeError):
                RESOURCE._get_relation_values(
                    RESOURCE.antonyms, term, lang=lang, sense=sense)


class TestAntonyms:

    def test_property(self):
        assert isinstance(RESOURCE.antonyms, dict)
        assert isinstance(RESOURCE.antonyms["en"], dict)
        assert isinstance(RESOURCE.antonyms["en"]["love"], dict)
        assert isinstance(RESOURCE.antonyms["en"]["love"]["v"], list)
        assert isinstance(RESOURCE.antonyms["en"]["love"]["v"][0], str)

    def test_method(self, spacy_doc):
        for tok in spacy_doc:
            assert isinstance(RESOURCE.get_antonyms(tok), list)
            assert isinstance(
                RESOURCE.get_antonyms(tok.text, lang="en", sense=tok.pos_), list)


class TestHyponyms:

    def test_property(self):
        assert isinstance(RESOURCE.hyponyms, dict)
        assert isinstance(RESOURCE.hyponyms["en"], dict)
        assert isinstance(RESOURCE.hyponyms["en"]["love"], dict)
        assert isinstance(RESOURCE.hyponyms["en"]["love"]["n"], list)
        assert isinstance(RESOURCE.hyponyms["en"]["love"]["n"][0], str)

    def test_method(self, spacy_doc):
        for tok in spacy_doc:
            assert isinstance(RESOURCE.get_hyponyms(tok), list)
            assert isinstance(
                RESOURCE.get_hyponyms(tok.text, lang="en", sense=tok.pos_), list)


class TestMeronyms:

    def test_property(self):
        assert isinstance(RESOURCE.meronyms, dict)
        assert isinstance(RESOURCE.meronyms["en"], dict)
        assert isinstance(RESOURCE.meronyms["en"]["ring"], dict)
        assert isinstance(RESOURCE.meronyms["en"]["ring"]["n"], list)
        assert isinstance(RESOURCE.meronyms["en"]["ring"]["n"][0], str)

    def test_method(self, spacy_doc):
        for tok in spacy_doc:
            assert isinstance(RESOURCE.get_meronyms(tok), list)
            assert isinstance(
                RESOURCE.get_meronyms(tok.text, lang="en", sense=tok.pos_), list)


class TestSynonyms:

    def test_property(self):
        assert isinstance(RESOURCE.synonyms, dict)
        assert isinstance(RESOURCE.synonyms["en"], dict)
        assert isinstance(RESOURCE.synonyms["en"]["love"], dict)
        assert isinstance(RESOURCE.synonyms["en"]["love"]["n"], list)
        assert isinstance(RESOURCE.synonyms["en"]["love"]["n"][0], str)

    def test_method(self, spacy_doc):
        for tok in spacy_doc:
            assert isinstance(RESOURCE.get_synonyms(tok), list)
            assert isinstance(
                RESOURCE.get_synonyms(tok.text, lang="en", sense=tok.pos_), list)



import datetime
import pathlib

import pytest

from textacy import utils


@pytest.mark.parametrize(
    "val,val_type,col_type,expected",
    [
        (None, int, list, None),
        (1, int, list, [1]),
        ([1, 2], int, tuple, (1, 2)),
        ((1, 1.0), (int, float), set, {1, 1.0}),
    ],
)
def test_to_collection(val, val_type, col_type, expected):
    assert utils.to_collection(val, val_type, col_type) == expected


class TestToUnicode:

    @pytest.mark.parametrize("s", [b"bytes", "unicode", "úñîçødé"])
    def test_valid(self, s):
        assert isinstance(utils.to_unicode(s), str)

    @pytest.mark.parametrize("s", [1, 2.0, ["foo", "bar"], {"foo": "bar"}])
    def test_invalid(self, s):
        with pytest.raises(TypeError):
            _ = utils.to_unicode(s)


class TestToBytes:

    @pytest.mark.parametrize("s", [b"bytes", "unicode", "úñîçødé"])
    def test_valid(self, s):
        assert isinstance(utils.to_bytes(s), bytes)

    @pytest.mark.parametrize("s", [1, 2.0, ["foo", "bar"], {"foo": "bar"}])
    def test_invalid(self, s):
        with pytest.raises(TypeError):
            _ = utils.to_bytes(s)


class TestToPath:

    @pytest.mark.parametrize("path", [pathlib.Path("."), pathlib.Path.home()])
    def test_path_input(self, path):
        assert isinstance(utils.to_path(path), pathlib.Path)

    @pytest.mark.parametrize("path", ["unicode", "úñîçødé"])
    def test_str_input(self, path):
        assert isinstance(utils.to_path(path), pathlib.Path)

    @pytest.mark.parametrize("path", [1, 2.0, ["foo", "bar"], {"foo": "bar"}])
    def test_invalid_input(self, path):
        with pytest.raises(TypeError):
            _ = utils.to_path(path)


class TestValidateAndClipRange:

    @pytest.mark.parametrize(
        "range_vals,full_range,val_type",
        [
            [("2001-01", "2002-01"), ("2000-01", "2003-01"), None],
            [["2001-01", "2004-01"], ("2000-01", "2003-01"), None],
            [("2001-01", "2002-01"), ["2000-01", "2003-01"], (str, bytes)],
            [[-5, 5], [-10, 10], None],
            [(-5, 5), (0, 10), None],
            [(-5, 5), (-10, 10), int],
            [(-5, 5), (-10, 10), (int, float)],
            [(0, None), (-5, 5), None],
            [(None, 0), (-5, 5), None],
        ],
    )
    def test_valid_inputs(self, range_vals, full_range, val_type):
        output = utils.validate_and_clip_range(range_vals, full_range, val_type)
        assert isinstance(output, tuple)
        assert len(output) == 2
        if range_vals[0] is None:
            assert output[0] == full_range[0]
        else:
            assert output[0] == max(range_vals[0], full_range[0])
        if range_vals[1] is None:
            assert output[1] == full_range[1]
        else:
            assert output[1] == min(range_vals[1], full_range[1])

    @pytest.mark.parametrize(
        "range_vals,full_range,val_type,error",
        [
            ["2001-01", ("2000-01", "2003-01"), None, pytest.raises(TypeError)],
            [("2001-01", "2002-01"), "2000-01", None, pytest.raises(TypeError)],
            [
                {"2001-01", "2002-01"},
                ("2000-01", "2003-01"),
                None,
                pytest.raises(TypeError),
            ],
            [
                ("2001-01", "2002-01"),
                ("2000-01", "2003-01"),
                datetime.date,
                pytest.raises(TypeError),
            ],
            [0, [-10, 10], None, pytest.raises(TypeError)],
            [(-5, 5), 0, None, pytest.raises(TypeError)],
            [[-5, 5], [-10, 10], (str, bytes), pytest.raises(TypeError)],
            [
                ("2001-01", "2002-01", "2003-01"),
                ("2000-01", "2003-01"),
                None,
                pytest.raises(ValueError),
            ],
            [
                ("2001-01", "2002-01"),
                ["2000-01", "2002-01", "2004-01"],
                None,
                pytest.raises(ValueError),
            ],
            [[0, 5, 10], (-10, 10), None, pytest.raises(ValueError)],
            [(-5, 5), [-10, 0, 10], None, pytest.raises(ValueError)],
            [(-5, 5), [-10, 0, 10], (str, bytes), pytest.raises(ValueError)],
        ],
    )
    def test_invalid_inputs(self, range_vals, full_range, val_type, error):
        with error:
            _ = utils.validate_and_clip_range(range_vals, full_range, val_type)


class TestValidateSetMembers:

    @pytest.mark.parametrize(
        "vals,val_type,valid_vals",
        [
            [{"a", "b"}, (str, bytes), {"a", "b", "c"}],
            ["a", (str, bytes), {"a", "b", "c"}],
            [("a", "b"), (str, bytes), {"a", "b", "c"}],
            [["a", "b"], (str, bytes), None],
            [{1, 2}, int, {1, 2, 3}],
            [{1, 2}, (int, float), {1, 2, 3}],
            [1, int, {1: "a", 2: "b", 3: "c"}],
            [{3.14, 42.0}, float, None],
            [3.14, (int, float), None],
        ]
    )
    def test_valid_inputs(self, vals, val_type, valid_vals):
        output = utils.validate_set_members(vals, val_type, valid_vals)
        assert isinstance(output, set)
        assert all(isinstance(val, val_type) for val in output)


    @pytest.mark.parametrize(
        "vals,val_type,valid_vals,error",
        [
            [{"a", "b"}, int, None, pytest.raises(TypeError)],
            ["a", int, None, pytest.raises(TypeError)],
            [("a", "b"), (int, float), None, pytest.raises(TypeError)],
            [{"a", "b"}, (str, bytes), {"x", "y", "z"}, pytest.raises(ValueError)],
            [{"a", "x"}, (str, bytes), {"x", "y", "z"}, pytest.raises(ValueError)],
            ["a", (str, bytes), {"x", "y", "z"}, pytest.raises(ValueError)],
            ["a", (str, bytes), {"x": 24, "y": 25, "z": 26}, pytest.raises(ValueError)],
        ]
    )
    def test_invalid_inputs(self, vals, val_type, valid_vals, error):
        with error:
            _ = utils.validate_set_members(vals, val_type, valid_vals)



# TODO: uncomment this when we're only supporting PY3.8+
# def _func_pos_only_args(parg1, parg2, /):
#     return (parg1, parg2)


# TODO: uncomment this when we're only supporting PY3.8+
# def _func_mix_args(parg, /, arg, *, kwarg):
#     return (parg, arg, kwarg)


def _func_mix_args(arg, *, kwarg):
    return (arg, kwarg)


def _func_kw_only_args(*, kwarg1, kwarg2):
    return (kwarg1, kwarg2)


@pytest.mark.parametrize(
    "func,kwargs,expected",
    [
        # (_func_pos_only_args, {"kwarg": "kwargval"}, {}),
        (_func_mix_args, {"arg": "argval"}, {"arg": "argval"}),
        (
            _func_mix_args,
            {"arg": "argval", "kwarg": "kwarval"},
            {"arg": "argval", "kwarg": "kwarval"},
        ),
        (
            _func_mix_args,
            {"arg": "argval", "kwarg": "kwargval", "foo": "bar"},
            {"arg": "argval", "kwarg": "kwargval"},
        ),
        (
            _func_kw_only_args,
            {"kwarg1": "kwarg1val", "kwarg2": "kwarg2val"},
            {"kwarg1": "kwarg1val", "kwarg2": "kwarg2val"},
        ),
        (
            _func_kw_only_args,
            {"kwarg1": "kwarg1val", "kwarg3": "kwarg3val"},
            {"kwarg1": "kwarg1val"},
        ),
        (_func_kw_only_args, {}, {}),
    ],
)
def test_get_kwargs_for_func(func, kwargs, expected):
    assert utils.get_kwargs_for_func(func, kwargs) == expected


@pytest.mark.parametrize(
    "text, n, pad, exp",
    [
        (
            "testing 123",
            1,
            False,
            ('t', 'e', 's', 't', 'i', 'n', 'g', ' ', '1', '2', '3'),
        ),
        (
            "testing 123",
            1,
            True,
            ('t', 'e', 's', 't', 'i', 'n', 'g', ' ', '1', '2', '3'),
        ),
        (
            "testing 123",
            2,
            False,
            ('te', 'es', 'st', 'ti', 'in', 'ng', 'g ', ' 1', '12', '23'),
        ),
        (
            "testing 123",
            2,
            True,
            ('_t', 'te', 'es', 'st', 'ti', 'in', 'ng', 'g ', ' 1', '12', '23', '3_'),
        ),
    ]
)
def test_text_to_char_ngrams(text, n, pad, exp):
    obs = utils.text_to_char_ngrams(text, n, pad=pad)
    assert all(isinstance(cng, str) and len(cng) == n for cng in obs)
    assert obs == exp



import os
import tarfile
import zipfile

import numpy as np
import pytest
from scipy import sparse as sp

from textacy import load_spacy_lang
from textacy import io, utils

TEXT = (
    "The year was 2081, and everybody was finally equal. "
    "They weren't only equal before God and the law. "
    "They were equal every which way."
)
TESTS_DIR = os.path.split(__file__)[0]


@pytest.fixture(scope="module")
def spacy_doc():
    spacy_lang = load_spacy_lang("en_core_web_sm")
    spacy_doc = spacy_lang(TEXT)
    return spacy_doc


class TestTextIO:

    def test_read_write_bytes(self, tmpdir):
        expected = utils.to_bytes(TEXT)
        for ext in (".txt", ".gz", ".bz2", ".xz"):
            filepath = str(tmpdir.join("test_read_write_file_bytes" + ext))
            io.write_text(expected, filepath, mode="wb", make_dirs=True)
            observed = next(io.read_text(filepath, mode="rb"))
            assert observed == expected

    def test_read_write_unicode(self, tmpdir):
        expected = TEXT
        for ext in (".txt", ".gz", ".bz2", ".xz"):
            filepath = str(tmpdir.join("test_read_write_file_unicode" + ext))
            io.write_text(expected, filepath, mode="wt", make_dirs=True)
            observed = next(io.read_text(filepath, mode="rt"))
            assert observed == expected

    def test_read_write_bytes_lines(self, tmpdir, spacy_doc):
        expected = [utils.to_bytes(sent.text) for sent in spacy_doc.sents]
        for ext in (".txt", ".gz", ".bz2", ".xz"):
            filepath = str(tmpdir.join("test_read_write_file_lines_bytes" + ext))
            io.write_text(expected, filepath, mode="wb", make_dirs=True, lines=True)
            observed = [
                line.strip() for line in io.read_text(filepath, mode="rb", lines=True)
            ]
            assert observed == expected

    def test_read_write_unicode_lines(self, tmpdir, spacy_doc):
        expected = [sent.text for sent in spacy_doc.sents]
        for ext in (".txt", ".gz", ".bz2", ".xz"):
            filepath = str(tmpdir.join("test_read_write_file_lines_unicode" + ext))
            io.write_text(expected, filepath, mode="wt", make_dirs=True, lines=True)
            observed = [
                line.strip() for line in io.read_text(filepath, mode="rt", lines=True)
            ]
            assert observed == expected


class TestJSONIO:

    def test_read_write_bytes(self, tmpdir, spacy_doc):
        expected = [{"idx": i, "sent": sent.text} for i, sent in enumerate(spacy_doc.sents)]
        for ext in (".json", ".json.gz", ".json.bz2", ".json.xz"):
            filepath = str(tmpdir.join("test_read_write_json_bytes" + ext))
            with pytest.raises(TypeError):
                io.write_json(expected, filepath, "wb", make_dirs=True)

    def test_read_write_unicode(self, tmpdir, spacy_doc):
        expected = [{"idx": i, "sent": sent.text} for i, sent in enumerate(spacy_doc.sents)]
        for ext in (".json", ".json.gz", ".json.bz2", ".json.xz"):
            filepath = str(tmpdir.join("test_read_write_json_unicode" + ext))
            io.write_json(expected, filepath, mode="wt", make_dirs=True)
            observed = next(io.read_json(filepath, mode="rt", lines=False))
            assert observed == expected

    def test_read_write_bytes_lines(self, tmpdir, spacy_doc):
        expected = [{"idx": i, "sent": sent.text} for i, sent in enumerate(spacy_doc.sents)]
        for ext in (".json", ".json.gz", ".json.bz2", ".json.xz"):
            filepath = str(tmpdir.join("test_read_write_json_lines_bytes" + ext))
            with pytest.raises(TypeError):
                io.write_json(
                    expected,
                    filepath,
                    mode="wb",
                    encoding=None,
                    make_dirs=True,
                    lines=True,
                )

    def test_read_write_unicode_lines(self, tmpdir, spacy_doc):
        expected = [{"idx": i, "sent": sent.text} for i, sent in enumerate(spacy_doc.sents)]
        for ext in (".json", ".json.gz", ".json.bz2", ".json.xz"):
            filepath = str(tmpdir.join("test_read_write_json_lines_unicode" + ext))
            io.write_json(expected, filepath, mode="wt", make_dirs=True, lines=True)
            observed = list(io.read_json(filepath, mode="rt", lines=True))
            assert observed == expected


class TestCSVIO:

    def test_read_write_compressed(self, tmpdir):
        expected = [
            ["this is some text", "scandal", 42.0],
            ["here's some more text: boom!", "escándalo", 1.0],
        ]
        for ext in (".csv", ".csv.gz", ".csv.bz2", ".csv.xz"):
            filepath = str(tmpdir.join("test_read_write_csv" + ext))
            io.write_csv(expected, filepath, make_dirs=True)
            observed = list(io.read_csv(filepath))
            assert observed == expected

    def test_read_write_delimiters(self, tmpdir):
        expected = [
            ["this is some text", "scandal", 42.0],
            ["here's some more text: boom!", "escándalo", 1.0],
        ]
        for delimiter in (",", "\t", "|", ":"):
            filepath = str(tmpdir.join("test_read_write_csv.csv"))
            io.write_csv(expected, filepath, delimiter=delimiter, make_dirs=True)
            observed = list(io.read_csv(filepath, delimiter=delimiter))
            assert observed == expected

    def test_read_write_dialect(self, tmpdir):
        expected = [
            ["this is some text", "scandal", 42.0],
            ["here's some more text: boom!", "escándalo", 1.0],
        ]
        filepath = str(tmpdir.join("test_read_write_csv.csv"))
        io.write_csv(expected, filepath, dialect="excel", make_dirs=True)
        observed = list(io.read_csv(filepath, dialect="infer"))
        assert observed == expected

    def test_read_write_dict(self, tmpdir):
        expected = [
            {"text": "this is some text", "kind": "scandal", "number": 42.0},
            {"text": "here's some more text: boom!", "kind": "escándalo", "number": 1.0},
        ]
        filepath = str(tmpdir.join("test_read_write_csv_dict.csv"))
        io.write_csv(
            expected,
            filepath,
            dialect="excel",
            make_dirs=True,
            fieldnames=["text", "kind", "number"],
        )
        observed = [
            dict(item)
            for item in io.read_csv(
                filepath, dialect="excel", fieldnames=["text", "kind", "number"]
            )
        ]
        assert observed == expected


class TestSpacyIO:

    @pytest.mark.skip(reason="this takes wayyy too long now, reason unknown")
    def test_read_write_docs(self, tmpdir, spacy_doc):
        expected = [tok.lower_ for tok in spacy_doc]
        for ext in (".pkl", ".pkl.gz", ".pkl.bz2", ".pkl.xz"):
            filepath = str(tmpdir.join("test_read_write_spacy_docs" + ext))
            io.write_spacy_docs(spacy_doc, filepath, make_dirs=True)
            observed = [
                tok.lower_ for doc in io.read_spacy_docs(filepath) for tok in doc
            ]
            assert observed == expected

    def test_read_write_docs_binary(self, tmpdir, spacy_doc):
        expected = [tok.lower_ for tok in spacy_doc]
        filepath = str(tmpdir.join("test_read_write_spacy_docs_binary.bin"))
        io.write_spacy_docs(spacy_doc, filepath, make_dirs=True, format="binary")
        with pytest.raises(ValueError):
            next(io.read_spacy_docs(filepath, format="binary", lang=None))
        observed = [
            tok.lower_
            for doc in io.read_spacy_docs(filepath, format="binary", lang="en_core_web_sm")
            for tok in doc
        ]
        assert observed == expected

    def test_read_write_docs_binary_attrs(self, tmpdir, spacy_doc):
        expected = [tok.tag_ for tok in spacy_doc]
        filepath = str(tmpdir.join("test_read_write_spacy_docs_binary_exclude.bin"))
        io.write_spacy_docs(
            spacy_doc, filepath, make_dirs=True,
            format="binary", attrs=["ORTH", "TAG"], store_user_data=False,
        )
        observed = [
            tok.tag_
            for doc in io.read_spacy_docs(filepath, format="binary", lang="en_core_web_sm")
            for tok in doc
        ]
        assert observed == expected


class TestMatrixIO:

    def test_read_write_sparse_csr(self, tmpdir):
        expected = sp.csr_matrix(
            (
                np.array([1, 2, 3, 4, 5, 6]),
                (np.array([0, 0, 1, 2, 2, 2]), np.array([0, 2, 2, 0, 1, 2])),
            ),
            shape=(3, 3),
        )
        filepath = str(tmpdir.join("test_read_write_sparse_matrix_csr.npz"))
        io.write_sparse_matrix(expected, filepath, compressed=False)
        observed = io.read_sparse_matrix(filepath, kind="csr")
        assert abs(observed - expected).nnz == 0

    def test_read_write_sparse_csr_compressed(self, tmpdir):
        expected = sp.csr_matrix(
            (
                np.array([1, 2, 3, 4, 5, 6]),
                (np.array([0, 0, 1, 2, 2, 2]), np.array([0, 2, 2, 0, 1, 2])),
            ),
            shape=(3, 3),
        )
        filepath = str(tmpdir.join("test_read_write_sparse_matrix_csr_compressed.npz"))
        io.write_sparse_matrix(expected, filepath, compressed=True)
        observed = io.read_sparse_matrix(filepath, kind="csr")
        assert abs(observed - expected).nnz == 0

    def test_read_write_sparse_csc(self, tmpdir):
        expected = sp.csc_matrix(
            (
                np.array([1, 2, 3, 4, 5, 6]),
                (np.array([0, 0, 1, 2, 2, 2]), np.array([0, 2, 2, 0, 1, 2])),
            ),
            shape=(3, 3),
        )
        filepath = str(tmpdir.join("test_read_write_sparse_matrix_csc.npz"))
        io.write_sparse_matrix(expected, filepath, compressed=False)
        observed = io.read_sparse_matrix(filepath, kind="csc")
        assert abs(observed - expected).nnz == 0

    def test_read_write_sparse_csc_compressed(self, tmpdir):
        expected = sp.csc_matrix(
            (
                np.array([1, 2, 3, 4, 5, 6]),
                (np.array([0, 0, 1, 2, 2, 2]), np.array([0, 2, 2, 0, 1, 2])),
            ),
            shape=(3, 3),
        )
        filepath = str(tmpdir.join("test_read_write_sparse_matrix_csc_compressed.npz"))
        io.write_sparse_matrix(expected, filepath, compressed=True)
        observed = io.read_sparse_matrix(filepath, kind="csc")
        assert abs(observed - expected).nnz == 0


class TestIOUtils:

    def test_get_filepaths(self):
        expected = sorted(
            os.path.join(TESTS_DIR, fname)
            for fname in os.listdir(TESTS_DIR)
            if os.path.isfile(os.path.join(TESTS_DIR, fname))
        )
        observed = sorted(
            io.get_filepaths(TESTS_DIR, ignore_invisible=False, recursive=False)
        )
        assert observed == expected

    def test_get_filepaths_ignore_invisible(self):
        dirpath = os.path.dirname(os.path.abspath(__file__))
        assert len(list(io.get_filepaths(dirpath, ignore_invisible=True))) <= len(
            list(io.get_filepaths(dirpath, ignore_invisible=False))
        )

    def test_get_filepaths_ignore_regex(self):
        assert (
            len(
                list(
                    io.get_filepaths(TESTS_DIR, ignore_regex="test_", ignore_invisible=True)
                )
            )
            == 0
        )

    def test_get_filepaths_match_regex(self):
        result = list(io.get_filepaths(TESTS_DIR, match_regex="_io", extension=".py"))
        assert len(result) == 1

    def test_get_filename_from_url(self):
        url_fnames = [
            ["http://www.foo.bar/bat.zip", "bat.zip"],
            ["www.foo.bar/bat.tar.gz", "bat.tar.gz"],
            ["foo.bar/bat.zip?q=test", "bat.zip"],
            ["http%3A%2F%2Fwww.foo.bar%2Fbat.tar.gz", "bat.tar.gz"]
        ]
        for url, fname in url_fnames:
            assert io.get_filename_from_url(url) == fname

    def test_unpack_archive(self, tmpdir):
        data = "Here's some text data to pack and unpack."
        fpath_txt = str(tmpdir.join("test_unpack_archive.txt"))
        with io.open_sesame(fpath_txt, mode="wt") as f:
            f.write(data)
        fpath_zip = str(tmpdir.join("test_unpack_archive.zip"))
        with zipfile.ZipFile(fpath_zip, "w") as f:
            f.write(fpath_txt)
        io.unpack_archive(fpath_zip, extract_dir=tmpdir)
        fpath_tar = str(tmpdir.join("test_unpack_archive.tar"))
        with tarfile.TarFile(fpath_tar, "w") as f:
            f.add(fpath_txt)
        io.unpack_archive(fpath_tar, extract_dir=tmpdir)
        io.unpack_archive(fpath_txt, extract_dir=tmpdir)



"""
Sparse Vectors
--------------

:mod:`textacy.representations.sparse_vec`: Transform a collection of tokenized docs
into a doc-term matrix of shape (# docs, # unique terms) or a group-term matrix
of shape (# unique groups, # unique terms), with various ways to filter/limit
included terms and flexible weighting/normalization schemes for their values.

Intended primarily as a simpler- and higher-level API for sparse vectorization of docs.
"""
from typing import Dict, Iterable, Optional, Tuple

import scipy.sparse as sp

from . import vectorizers


def build_doc_term_matrix(
    tokenized_docs: Iterable[Iterable[str]],
    *,
    tf_type: str = "linear",  # Literal["linear", "sqrt", "log", "binary"]
    idf_type: Optional[str] = None,  # Optional[Literal["standard", "smooth", "bm25"]]
    dl_type: Optional[str] = None,  # Optional[Literal["linear", "sqrt", "log"]]
    **kwargs
) -> Tuple[sp.csr_matrix, Dict[str, int]]:
    """
    Transform one or more tokenized documents into a document-term matrix
    of shape (# docs, # unique terms), with flexible weighting/normalization of values.

    Args:
        tokenized_docs: A sequence of tokenized documents, where each is a sequence
            of term strings. For example::

                >>> ([tok.lemma_ for tok in spacy_doc]
                ...  for spacy_doc in spacy_docs)
                >>> ((ne.text for ne in extract.entities(doc))
                ...  for doc in corpus)

        tf_type: Type of term frequency (tf) to use for weights' local component:

            - "linear": tf (tfs are already linear, so left as-is)
            - "sqrt": tf => sqrt(tf)
            - "log": tf => log(tf) + 1
            - "binary": tf => 1

        idf_type: Type of inverse document frequency (idf) to use for weights'
            global component:

            - "standard": idf = log(n_docs / df) + 1.0
            - "smooth": idf = log(n_docs + 1 / df + 1) + 1.0, i.e. 1 is added
              to all document frequencies, as if a single document containing
              every unique term was added to the corpus.
            - "bm25": idf = log((n_docs - df + 0.5) / (df + 0.5)), which is
              a form commonly used in information retrieval that allows for
              very common terms to receive negative weights.
            - None: no global weighting is applied to local term weights.

        dl_type: Type of document-length scaling to use for weights'
            normalization component:

            - "linear": dl (dls are already linear, so left as-is)
            - "sqrt": dl => sqrt(dl)
            - "log": dl => log(dl)
            - None: no normalization is applied to local(*global?) weights

        **kwargs: Passed directly into vectorizer class

    Returns:
        Document-term matrix as a sparse row matrix, and
        the corresponding mapping of term strings to integer ids (column indexes).

    Note:
        If you need to transform other sequences of tokenized documents in the same way,
        or if you need more access to the underlying vectorization process,
        consider using :class:`textacy.representations.vectorizers.Vectorizer` directly.

    See Also:
        - :class:`textacy.representations.vectorizers.Vectorizer`
        - :class:`scipy.sparse.csr_matrix`

    Reference:
        https://en.wikipedia.org/wiki/Document-term_matrix
    """
    vectorizer = vectorizers.Vectorizer(
        tf_type=tf_type, idf_type=idf_type, dl_type=dl_type, **kwargs
    )
    doc_term_matrix = vectorizer.fit_transform(tokenized_docs)
    return (doc_term_matrix, vectorizer.vocabulary_terms)


def build_grp_term_matrix(
    tokenized_docs: Iterable[Iterable[str]],
    grps: Iterable[str],
    *,
    tf_type: str = "linear",  # Literal["linear", "sqrt", "log", "binary"]
    idf_type: Optional[str] = None,  # Optional[Literal["standard", "smooth", "bm25"]]
    dl_type: Optional[str] = None,  # Optional[Literal["linear", "sqrt", "log"]]
    **kwargs
) -> sp.csr_matrix:
    """
    Transform one or more tokenized documents into a group-term matrix
    of shape (# unique groups, # unique terms),
    with flexible weighting/normalization of values.

    This is an extension of typical document-term matrix vectorization, where
    terms are grouped by the documents in which they co-occur. It allows for
    customized grouping, such as by a shared author or publication year, that
    may span multiple documents, without forcing users to merge those documents
    themselves.

    Args:
        tokenized_docs: A sequence of tokenized documents, where each is a sequence
            of term strings. For example::

                >>> ([tok.lemma_ for tok in spacy_doc]
                ...  for spacy_doc in spacy_docs)
                >>> ((ne.text for ne in extract.entities(doc))
                ...  for doc in corpus)

        grps: Sequence of group names by which the terms in ``tokenized_docs``
            are aggregated, where the first item in ``grps`` corresponds to
            the first item in ``tokenized_docs``, and so on.
        tf_type: Type of term frequency (tf) to use for weights' local component:

            - "linear": tf (tfs are already linear, so left as-is)
            - "sqrt": tf => sqrt(tf)
            - "log": tf => log(tf) + 1
            - "binary": tf => 1

        idf_type: Type of inverse document frequency (idf) to use for weights'
            global component:

            - "standard": idf = log(n_docs / df) + 1.0
            - "smooth": idf = log(n_docs + 1 / df + 1) + 1.0, i.e. 1 is added
              to all document frequencies, as if a single document containing
              every unique term was added to the corpus.
            - "bm25": idf = log((n_docs - df + 0.5) / (df + 0.5)), which is
              a form commonly used in information retrieval that allows for
              very common terms to receive negative weights.
            - None: no global weighting is applied to local term weights.

        dl_type: Type of document-length scaling to use for weights'
            normalization component:

            - "linear": dl (dls are already linear, so left as-is)
            - "sqrt": dl => sqrt(dl)
            - "log": dl => log(dl)
            - None: no normalization is applied to local(*global?) weights

        **kwargs: Passed directly into vectorizer class

    Returns:
        Group-term matrix as a sparse row matrix, and
        the corresponding mapping of term strings to integer ids (column indexes), and
        the corresponding mapping of group strings to integer ids (row indexes).

    Note:
        If you need to transform other sequences of tokenized documents in the same way,
        or if you need more access to the underlying vectorization process, consider
        using :class:`textacy.representations.vectorizers.GroupVectorizer` directly.

    See Also:
        - :class:`textacy.representations.vectorizers.GroupVectorizer`
        - :class:`scipy.sparse.csr_matrix`

    Reference:
        https://en.wikipedia.org/wiki/Document-term_matrix
    """
    vectorizer = vectorizers.GroupVectorizer(
        tf_type=tf_type, idf_type=idf_type, dl_type=dl_type, **kwargs
    )
    grp_term_matrix = vectorizer.fit_transform(tokenized_docs, grps)
    return (grp_term_matrix, vectorizer.vocabulary_terms, vectorizer.vocabulary_grps)



import pytest

import textacy
from textacy.extract import utils


@pytest.fixture(scope="module")
def doc():
    lang = textacy.load_spacy_lang("en_core_web_sm")
    text = (
        "Many years later, as he faced the firing squad, Colonel Aureliano Buendía was "
        "to remember that distant afternoon when his father took him to discover ice. "
        "At that time Macondo was a village of twenty adobe houses, built on the bank "
        "of a river of clear water that ran along a bed of polished stones, which were "
        "white and enormous, like prehistoric eggs. The world was so recent that many "
        "things lacked names, and in order to indicate them it was necessary to point."
    )
    return textacy.make_spacy_doc(text, lang=lang)


@pytest.fixture(scope="module")
def term_tokens(doc):
    return list(textacy.extract.words(doc))


@pytest.fixture(scope="module")
def term_spans(doc):
    return list(textacy.extract.ngrams(doc, 2))


class TestTermsToStrings:

    def test_term_spans(self, term_spans):
        results = list(utils.terms_to_strings(term_spans, "orth"))
        assert results
        assert isinstance(results[0], str)

    def test_term_tokens(self, term_tokens):
        results = list(utils.terms_to_strings(term_tokens, "orth"))
        assert results
        assert isinstance(results[0], str)

    @pytest.mark.parametrize("by", ["orth", "lower", "lemma"])
    def test_by_str(self, by, term_spans):
        results = list(utils.terms_to_strings(term_spans, by))
        assert results
        assert isinstance(results[0], str)

    @pytest.mark.parametrize("by", [lambda term: term.text])
    def test_by_callable(self, by, term_spans):
        results = list(utils.terms_to_strings(term_spans, by))
        assert results
        assert isinstance(results[0], str)

    @pytest.mark.parametrize("by", ["LEMMA", None, True])
    def test_by_invalid(self, by, term_spans):
        with pytest.raises(ValueError):
            _ = list(utils.terms_to_strings(term_spans, by))


def test_aggregate_term_variants():
    # TODO: the actual results are NOT what i'd expect; figure out why
    terms = set([
        "vice versa",
        "vice-versa",
        "vice/versa",
        "BJD",
        "Burton Jacob DeWilde",
        "the big black cat named Rico",
        "the black cat named Rico",
    ])
    result1 = utils.aggregate_term_variants(terms)
    result2 = utils.aggregate_term_variants(
        terms, acro_defs={"BJD": "Burton Jacob DeWilde"})
    assert len(result2) <= len(result1) <= len(terms)


# def test_most_discriminating_terms():
#     text1 = """Friedman joined the London bureau of United Press International after completing his master's degree. He was dispatched a year later to Beirut, where he lived from June 1979 to May 1981 while covering the Lebanon Civil War. He was hired by The New York Times as a reporter in 1981 and re-dispatched to Beirut at the start of the 1982 Israeli invasion of Lebanon. His coverage of the war, particularly the Sabra and Shatila massacre, won him the Pulitzer Prize for International Reporting (shared with Loren Jenkins of The Washington Post). Alongside David K. Shipler he also won the George Polk Award for foreign reporting.
#     In June 1984, Friedman was transferred to Jerusalem, where he served as the New York Times Jerusalem Bureau Chief until February 1988. That year he received a second Pulitzer Prize for International Reporting, which cited his coverage of the First Palestinian Intifada. He wrote a book, From Beirut to Jerusalem, describing his experiences in the Middle East, which won the 1989 U.S. National Book Award for Nonfiction.
#     Friedman covered Secretary of State James Baker during the administration of President George H. W. Bush. Following the election of Bill Clinton in 1992, Friedman became the White House correspondent for the New York Times. In 1994, he began to write more about foreign policy and economics, and moved to the op-ed page of The New York Times the following year as a foreign affairs columnist. In 2002, Friedman won the Pulitzer Prize for Commentary for his "clarity of vision, based on extensive reporting, in commenting on the worldwide impact of the terrorist threat."
#     In February 2002, Friedman met Saudi Crown Prince Abdullah and encouraged him to make a comprehensive attempt to end the Arab-Israeli conflict by normalizing Arab relations with Israel in exchange for the return of refugees alongside an end to the Israel territorial occupations. Abdullah proposed the Arab Peace Initiative at the Beirut Summit that March, which Friedman has since strongly supported.
#     Friedman received the 2004 Overseas Press Club Award for lifetime achievement and was named to the Order of the British Empire by Queen Elizabeth II.
#     In May 2011, The New York Times reported that President Barack Obama "has sounded out" Friedman concerning Middle East issues."""

#     text2 = """In 1954, Bucksbaum and his brother Martin borrowed $1.2 million and built the first shopping center in Cedar Rapids, Iowa, anchored by a fourth family grocery store.
#     They expanded into enclosed malls which mirrored the continued movement to the suburbs seen in the 1960s.
#     By 1964, their company - then named General Management - owned five malls anchored by the Younkers department store.
#     In 1972, the company became publicly traded on the New York Stock Exchange under the name General Growth Properties and became the second-largest owner, developer, and manager of regional shopping malls in the country.
#     Bucksbaum served as its Chairman and Chief Executive Officer and under his tenure, he formed two Real estate investment trusts and expanded the company's portfolio of malls and shopping centers via more than $36 billion in acquisitions.
#     In 1984, General Growth sold 19 malls for $800 million to Equitable Real Estate, which was deemed the “nation’s largest single asset real estate transaction” to date.
#     In 1995, his brother Martin died and he re-located the company to Chicago.
#     In 2004, General Growth purchased The Rouse Company for $14.2 billion.
#     By 2007, General Growth was the second-largest REIT owning 194 malls with over 200 million square feet in 44 states.
#     In 2008, General Growth filed for Chapter 11 bankruptcy protection after the collapse of the stock market."""

#     doc1 = [line.split() for line in text1.split("\n")]
#     doc2 = [line.split() for line in text2.split("\n")]

#     expected = (["Friedman", "Times"], ["General", "malls"])
#     observed = utils.most_discriminating_terms(
#         doc1 + doc2, [True] * len(doc1) + [False] * len(doc2), top_n_terms=2
#     )
#     assert expected == observed


@pytest.mark.parametrize(
    "input_,output_",
    [
        (
            ["foo (bar)", "foo?", "bar!", "-123.4"],
            ["foo (bar)", "foo?", "bar!", "-123.4"],
        ),
        (["(foo bar", "foo) bar", "?>,!-.", "", "foo) (bar"], []),
        (
            [
                "( foo bar )",
                "foo -bar",
                "- 123.4",
                ".-foo bar",
                "?!foo",
                "bar?!",
                "foo 's bar",
                "foo 'll bar",
                "  foo   bar   ",
                "foo bar.   ",
            ],
            [
                "(foo bar)",
                "foo-bar",
                "-123.4",
                "foo bar",
                "foo",
                "bar?!",
                "foo's bar",
                "foo'll bar",
                "foo bar",
                "foo bar.",
            ],
        ),
    ],
)
def test_clean_term_strings(input_, output_):
    assert list(utils.clean_term_strings(input_)) == output_



from functools import partial

import pytest

from textacy import preprocessing


@pytest.mark.parametrize(
    "funcs, text_in, text_out",
    [
        (
            [
                preprocessing.replace.hashtags,
                preprocessing.replace.user_handles,
                preprocessing.replace.emojis,
            ],
            "@spacy_io is OSS for industrial-strength NLP in Python developed by @explosion_ai 💥",
            "_USER_ is OSS for industrial-strength NLP in Python developed by _USER_ _EMOJI_",
        ),
        (
            [
                lambda x: x.lower(),
                preprocessing.replace.numbers,
                partial(preprocessing.remove.punctuation, only="."),
            ],
            "spaCy 3.0 introduces transformer-based pipelines for state-of-the-art performance.",
            "spacy _NUMBER_ introduces transformer-based pipelines for state-of-the-art performance ",
        ),
        # note: func order matters!
        (
            [
                preprocessing.replace.numbers,
                partial(preprocessing.remove.punctuation, only="."),
                lambda x: x.lower(),
            ],
            "spaCy 3.0 introduces transformer-based pipelines for state-of-the-art performance.",
            "spacy _number_ introduces transformer-based pipelines for state-of-the-art performance ",
        ),
    ]
)
def test_make_pipeline(funcs, text_in, text_out):
    pipeline = preprocessing.make_pipeline(*funcs)
    assert pipeline(text_in) == text_out



import datetime
import os

import pytest

from textacy.datasets.supreme_court import SupremeCourt

DATASET = SupremeCourt()

pytestmark = pytest.mark.skipif(
    DATASET.filepath is None,
    reason="SupremeCourt dataset must be downloaded before running tests",
)


@pytest.mark.skip("No need to download a new dataset every time")
def test_download(tmpdir):
    dataset = SupremeCourt(data_dir=str(tmpdir))
    dataset.download()
    assert os.path.isfile(dataset._filepath)


def test_oserror(tmpdir):
    dataset = SupremeCourt(data_dir=str(tmpdir))
    with pytest.raises(OSError):
        _ = list(dataset.texts())


def test_texts():
    for text in DATASET.texts(limit=3):
        assert isinstance(text, str)


def test_texts_limit():
    for limit in (1, 5, 10):
        assert sum(1 for _ in DATASET.texts(limit=limit)) == limit


def test_texts_min_len():
    for min_len in (100, 200, 1000):
        assert all(
            len(text) >= min_len for text in DATASET.texts(min_len=min_len, limit=10)
        )


def test_records():
    for text, meta in DATASET.records(limit=3):
        assert isinstance(text, str)
        assert isinstance(meta, dict)


def test_records_opinion_author():
    opinion_authors = ({78}, {78, 81})
    for opinion_author in opinion_authors:
        records = list(DATASET.records(opinion_author=opinion_author, limit=10))
        assert len(records) >= 1
        assert all(
            meta["maj_opinion_author"] in opinion_author
            for text, meta in records
        )


def test_records_decision_direction():
    decision_directions = ("liberal", {"conservative", "unspecifiable"})
    for decision_direction in decision_directions:
        records = list(DATASET.records(decision_direction=decision_direction, limit=10))
        assert len(records) >= 1
        assert all(
            meta["decision_direction"] in decision_direction
            for text, meta in records
        )


def test_records_issue_area():
    issue_areas = ({2}, {4, 5, 6})
    for issue_area in issue_areas:
        records = list(DATASET.records(issue_area=issue_area, limit=10))
        assert len(records) >= 1
        assert all(
            meta["issue_area"] in issue_area
            for text, meta in records
        )


def test_records_date_range():
    date_ranges = (["1947-01-01", "1948-01-01"], ("1947-07-01", "1947-12-31"))
    for date_range in date_ranges:
        records = list(DATASET.records(date_range=date_range, limit=10))
        assert len(records) >= 1
        assert all(
            date_range[0] <= meta["decision_date"] < date_range[1]
            for text, meta in records
        )


def test_bad_filters():
    bad_filters = (
        {"opinion_author": 1000},
        {"decision_direction": "blatantly political"},
        {"issue_area": 1000},
        {"min_len": -1},
    )
    for bad_filter in bad_filters:
        with pytest.raises(ValueError):
            list(DATASET.texts(**bad_filter))
    bad_filters = (
        {"opinion_author": "Burton DeWilde"},
        {"issue_area": "legalizing gay marriage, woo!"},
        {"date_range": "2016-01-01"},
        {"date_range": (datetime.date(2000, 1, 1), datetime.date(2001, 1, 1))},
    )
    for bad_filter in bad_filters:
        with pytest.raises(TypeError):
            list(DATASET.texts(**bad_filter))



import numpy as np

from .. import errors
from ..utils import to_collection

try:
    import matplotlib.pyplot as plt
    import matplotlib.transforms

except ImportError:
    pass


RC_PARAMS = {
    "axes.axisbelow": True,
    "axes.edgecolor": ".8",
    "axes.facecolor": "white",
    "axes.grid": True,
    "axes.labelcolor": ".15",
    "axes.linewidth": 1.0,
    "figure.facecolor": "white",
    "font.family": ["sans-serif"],
    "font.sans-serif": ["Arial", "Liberation Sans", "sans-serif"],
    "grid.color": ".8",
    "grid.linestyle": "-",
    "image.cmap": "Greys",
    "legend.frameon": False,
    "legend.numpoints": 1,
    "legend.scatterpoints": 1,
    "lines.solid_capstyle": "round",
    "text.color": ".15",
    "xtick.color": ".15",
    "xtick.direction": "out",
    "xtick.major.size": 0.0,
    "xtick.minor.size": 0.0,
    "ytick.color": ".15",
    "ytick.direction": "out",
    "ytick.major.size": 0.0,
    "ytick.minor.size": 0.0,
}

COLOR_PAIRS = (
    (
        (0.65098041296005249, 0.80784314870834351, 0.89019608497619629),
        (0.12572087695201239, 0.47323337360924367, 0.707327968232772),
    ),
    (
        (0.68899655751153521, 0.8681737867056154, 0.54376011946622071),
        (0.21171857311445125, 0.63326415104024547, 0.1812226118410335),
    ),
    (
        (0.98320646005518297, 0.5980161709820524, 0.59423301088459368),
        (0.89059593116535862, 0.10449827132271793, 0.11108035462744099),
    ),
    (
        (0.99175701702342312, 0.74648213716698619, 0.43401768935077328),
        (0.99990772780250103, 0.50099192647372981, 0.0051211073118098693),
    ),
    (
        (0.78329874347238004, 0.68724338552531095, 0.8336793640080622),
        (0.42485198495434734, 0.2511495584950722, 0.60386007743723258),
    ),
    (
        (0.99760092286502611, 0.99489427150464516, 0.5965244373854468),
        (0.69411766529083252, 0.3490196168422699, 0.15686275064945221),
    ),
)


def draw_termite_plot(
    values_mat,
    col_labels,
    row_labels,
    *,
    highlight_cols=None,
    highlight_colors=None,
    save=False,
    rc_params=None,
):
    """
    Make a "termite" plot, typically used for assessing topic models with a tabular
    layout that promotes comparison of terms both within and across topics.

    Args:
        values_mat (:class:`np.ndarray` or matrix): matrix of values with shape
            (# row labels, # col labels) used to size the dots on the grid
        col_labels (seq[str]): labels used to identify x-axis ticks on the grid
        row_labels(seq[str]): labels used to identify y-axis ticks on the grid
        highlight_cols (int or seq[int], optional): indices for columns
            to visually highlight in the plot with contrasting colors
        highlight_colors (tuple of 2-tuples): each 2-tuple corresponds to a pair
            of (light/dark) matplotlib-friendly colors used to highlight a single
            column; if not specified (default), a good set of 6 pairs are used
        save (str, optional): give the full /path/to/fname on disk to save figure
        rc_params (dict, optional): allow passing parameters to rc_context in matplotlib.plyplot,
            details in https://matplotlib.org/3.1.0/api/_as_gen/matplotlib.pyplot.rc_context.html

    Returns:
        :obj:`matplotlib.axes.Axes.axis`: Axis on which termite plot is plotted.

    Raises:
        ValueError: if more columns are selected for highlighting than colors
            or if any of the inputs' dimensions don't match

    References:
        Chuang, Jason, Christopher D. Manning, and Jeffrey Heer. "Termite:
        Visualization techniques for assessing textual topic models."
        Proceedings of the International Working Conference on Advanced
        Visual Interfaces. ACM, 2012.

    See Also:
        :meth:`TopicModel.termite_plot() <textacy.tm.topic_model.TopicModel.termite_plot>`
    """
    try:
        plt
    except NameError:
        raise ImportError(
            "`matplotlib` is not installed, so `textacy.viz` won't work; "
            "install it individually via `$ pip install matplotlib`, or "
            "along with textacy via `pip install textacy[viz]`."
        )
    n_rows, n_cols = values_mat.shape
    max_val = np.max(values_mat)

    if n_rows != len(row_labels):
        raise ValueError(
            "values_mat and row_labels dimensions don't match: "
            f"{n_rows} vs. {len(row_labels)}"
        )
    if n_cols != len(col_labels):
        raise ValueError(
            "values_mat and col_labels dimensions don't match: "
            f"{n_cols} vs. {len(col_labels)}"
        )

    if highlight_colors is None:
        highlight_colors = COLOR_PAIRS
    if highlight_cols is not None:
        if isinstance(highlight_cols, int):
            highlight_cols = (highlight_cols,)
        elif len(highlight_cols) > len(highlight_colors):
            raise ValueError(
                f"no more than {len(highlight_colors)} columns "
                "may be highlighted at once"
            )

        highlight_colors = {hc: COLOR_PAIRS[i] for i, hc in enumerate(highlight_cols)}

    _rc_params = RC_PARAMS.copy()
    if rc_params:
        _rc_params.update(rc_params)

    with plt.rc_context(RC_PARAMS):
        fig, ax = plt.subplots(figsize=(pow(n_cols, 0.8), pow(n_rows, 0.66)))

        _ = ax.set_yticks(range(n_rows))
        yticklabels = ax.set_yticklabels(row_labels, fontsize=14, color="gray")
        if highlight_cols is not None:
            for i, ticklabel in enumerate(yticklabels):
                max_tick_val = max(values_mat[i, hc] for hc in range(n_cols))
                for hc in highlight_cols:
                    if max_tick_val > 0 and values_mat[i, hc] == max_tick_val:
                        ticklabel.set_color(highlight_colors[hc][1])

        ax.get_xaxis().set_ticks_position("top")
        _ = ax.set_xticks(range(n_cols))
        xticklabels = ax.set_xticklabels(
            col_labels, fontsize=14, color="gray", rotation=-60, ha="right"
        )

        # Create offset transform by 5 points in x direction
        dx = 10 / 72
        dy = 0
        offset = matplotlib.transforms.ScaledTranslation(dx, dy, fig.dpi_scale_trans)
        for label in ax.xaxis.get_majorticklabels():
            label.set_transform(label.get_transform() + offset)

        if highlight_cols is not None:
            gridlines = ax.get_xgridlines()
            for i, ticklabel in enumerate(xticklabels):
                if i in highlight_cols:
                    ticklabel.set_color(highlight_colors[i][1])
                    gridlines[i].set_color(highlight_colors[i][0])
                    gridlines[i].set_alpha(0.5)

        for col_ind in range(n_cols):
            if highlight_cols is not None and col_ind in highlight_cols:
                ax.scatter(
                    [col_ind for _ in range(n_rows)],
                    [i for i in range(n_rows)],
                    s=600 * (values_mat[:, col_ind] / max_val),
                    alpha=0.5,
                    linewidth=1,
                    color=highlight_colors[col_ind][0],
                    edgecolor=highlight_colors[col_ind][1],
                )
            else:
                ax.scatter(
                    [col_ind for _ in range(n_rows)],
                    [i for i in range(n_rows)],
                    s=600 * (values_mat[:, col_ind] / max_val),
                    alpha=0.5,
                    linewidth=1,
                    color="lightgray",
                    edgecolor="gray",
                )

            _ = ax.set_xlim(left=-1, right=n_cols)
            _ = ax.set_ylim(bottom=-1, top=n_rows)

            ax.invert_yaxis()  # otherwise, values/labels go from bottom to top

    if save:
        fig.savefig(save, bbox_inches="tight", dpi=100)

    return ax


def termite_df_plot(
    components,
    *,
    highlight_topics=None,
    n_terms=25,
    rank_terms_by="max",
    sort_terms_by="seriation",
    save=False,
    rc_params=None,
):
    """
    Make a "termite" plot for assessing topic models using a tabular layout
    to promote comparison of terms both within and across topics.

    Args:
        components (:class:`pandas.DataFrame` or sparse matrix): corpus
            represented as a term-topic matrix with shape (n_terms, n_topics);
            should have terms as index and topics as column names
        topics (int or Sequence[int]): topic(s) to include in termite plot;
            if -1, all topics are included
        highlight_topics (str or Sequence[str]): names for up to 6 topics
            to visually highlight in the plot with contrasting colors
        n_terms (int): number of top terms to include in termite plot
        rank_terms_by ({'max', 'mean', 'var'}): argument to dataframe `agg`
            function, used to rank terms;
            the top-ranked ``n_terms`` are included in the plot
        sort_terms_by ({'seriation', 'weight', 'index', 'alphabetical'}):
            method used to vertically sort the selected top ``n_terms`` terms;
            the default ("seriation") groups similar terms together, which
            facilitates cross-topic assessment
        save (str): give the full /path/to/fname on disk to save figure
            rc_params (dict, optional): allow passing parameters to rc_context
            in matplotlib.plyplot, details in
            https://matplotlib.org/3.1.0/api/_as_gen/matplotlib.pyplot.rc_context.html

    Returns:
        ``matplotlib.axes.Axes.axis``: Axis on which termite plot is plotted.

    Raises:
        ValueError: if more than 6 topics are selected for highlighting, or
            an invalid value is passed for the sort_topics_by, rank_terms_by,
            and/or sort_terms_by params

    References:
        - Chuang, Jason, Christopher D. Manning, and Jeffrey Heer. "Termite:
            Visualization techniques for assessing textual topic models."
            Proceedings of the International Working Conference on Advanced
            Visual Interfaces. ACM, 2012.
        - Fajwel Fogel, Alexandre d’Aspremont, and Milan Vojnovic. 2016.
            Spectral ranking using seriation. J. Mach. Learn. Res. 17, 1
            (January 2016), 3013–3057.

    See Also:
        :func:`viz.termite_plot <textacy.viz.termite.termite_plot>`
    TODO: `rank_terms_by` other metrics, e.g. topic salience or relevance
    """
    highlight_topics = to_collection(highlight_topics, str, list)

    if len(highlight_topics) > 6:
        raise ValueError("no more than 6 topics may be highlighted at once")

    # get column index of any topics to highlight in termite plot
    if highlight_topics is not None:
        highlight_cols = tuple(components.columns.get_loc(c) for c in highlight_topics)
    else:
        highlight_cols = None

    component_filter = components.loc[
        components.agg(rank_terms_by, axis=1)
        .sort_values(ascending=False)
        .iloc[:n_terms]
        .index
    ]

    # get top term indices in sorted order
    if sort_terms_by == "weight":
        pass  # already done
    elif sort_terms_by == "alphabetical":
        component_filter = component_filter.sort_index()
    elif sort_terms_by == "seriation":
        # calculate similarity matrix
        similarity = (
            component_filter @ component_filter.T.pipe(lambda df: df - df.min().min())
        ).values

        # compute Laplacian matrice and its 2nd eigenvector
        L = np.diag(similarity.sum(axis=1)) - similarity
        D, V = np.linalg.eigh(L)
        D = D[np.argsort(D)]
        V = V[:, np.argsort(D)]
        fiedler = V[:, 1]

        # get permutation corresponding to sorting the 2nd eigenvector
        component_filter = component_filter.reindex(
            index=[component_filter.index[i] for i in np.argsort(fiedler)],
        )
    else:
        raise ValueError(
            errors.value_invalid_msg(
                "sort_terms_by", sort_terms_by, {"weight", "alphabetical", "seriation"},
            )
        )

    # get topic and term labels
    topic_labels = component_filter.columns
    term_labels = component_filter.index

    # get topic-term weights to size dots
    term_topic_weights = component_filter.values

    return draw_termite_plot(
        term_topic_weights,
        topic_labels,
        term_labels,
        highlight_cols=highlight_cols,
        save=save,
        rc_params=rc_params,
    )



"""
Remove
------

:mod:`textacy.preprocessing.remove`: Remove aspects of raw text that may be unwanted
for certain use cases.
"""
from __future__ import annotations

import re
import unicodedata
from typing import Collection, Optional

from . import resources
from .. import utils


def accents(text: str, *, fast: bool = False) -> str:
    """
    Remove accents from any accented unicode characters in ``text``, either by
    replacing them with ASCII equivalents or removing them entirely.

    Args:
        text
        fast: If False, accents are removed from any unicode symbol
            with a direct ASCII equivalent; if True, accented chars
            for all unicode symbols are removed, regardless.

            .. note:: ``fast=True`` can be significantly faster than ``fast=False``,
               but its transformation of ``text`` is less "safe" and more likely
               to result in changes of meaning, spelling errors, etc.

    Returns:
        str

    See Also:
        For a more powerful (but slower) alternative, check out ``unidecode``:
        https://github.com/avian2/unidecode
    """
    if fast is False:
        return "".join(
            char
            for char in unicodedata.normalize("NFKD", text)
            if not unicodedata.combining(char)
        )
    else:
        return (
            unicodedata.normalize("NFKD", text)
            .encode("ascii", errors="ignore")
            .decode("ascii")
        )


def brackets(
    text: str,
    *,
    only: Optional[str | Collection[str]] = None,
) -> str:
    """
    Remove text within curly {}, square [], and/or round () brackets, as well as
    the brackets themselves.

    Args:
        text
        only: Remove only those bracketed contents as specified here: "curly", "square",
            and/or "round". For example, ``"square"`` removes only those contents found
            between square brackets, while ``["round", "square"]`` removes those contents
            found between square or round brackets, but not curly.

    Returns:
        str

    Note:
        This function relies on regular expressions, applied sequentially for curly,
        square, then round brackets; as such, it doesn't handle nested brackets of the
        same type and may behave unexpectedly on text with "wild" use of brackets.
        It should be fine removing structured bracketed contents, as is often used,
        for instance, to denote in-text citations.
    """
    only = utils.to_collection(only, val_type=str, col_type=set)
    if only is None or "curly" in only:
        text = resources.RE_BRACKETS_CURLY.sub("", text)
    if only is None or "square" in only:
        text = resources.RE_BRACKETS_SQUARE.sub("", text)
    if only is None or "round" in only:
        text = resources.RE_BRACKETS_ROUND.sub("", text)
    return text


def html_tags(text: str) -> str:
    """
    Remove HTML tags from ``text``, returning just the text found between tags
    and other non-data elements.

    Args:
        text

    Returns:
        str

    Note:
        This function relies on the stdlib :class:`html.parser.HTMLParser` and
        doesn't do anything fancy. For a better and potentially faster solution,
        consider using ``lxml`` and/or ``beautifulsoup4``.
    """
    parser = resources.HTMLTextExtractor()
    parser.feed(text)
    return parser.get_text()


def punctuation(
    text: str,
    *,
    only: Optional[str | Collection[str]] = None,
) -> str:
    """
    Remove punctuation from ``text`` by replacing all instances of punctuation
    (or a subset thereof specified by ``only``) with whitespace.

    Args:
        text
        only: Remove only those punctuation marks specified here. For example,
            ``"."`` removes only periods, while ``[",", ";", ":"]`` removes commas,
            semicolons, and colons; if None, all unicode punctuation marks are removed.

    Returns:
        str

    Note:
        When ``only=None``, Python's built-in :meth:`str.translate()` is
        used to remove punctuation; otherwise, a regular expression is used.
        The former's performance can be up to an order of magnitude faster.
    """
    if only is not None:
        only = utils.to_collection(only, val_type=str, col_type=set)
        return re.sub("[{}]+".format(re.escape("".join(only))), " ", text)
    else:
        return text.translate(resources.PUNCT_TRANSLATION_TABLE)



from __future__ import annotations

import operator
from functools import partial
from typing import Callable, Collection, Iterable, Optional, Tuple

from cytoolz import itertoolz
from spacy.tokens import Span

from .. import extract, types


DocLikeToSpans = Callable[[types.DocLike], Iterable[Span]]


class TermsTokenizer:
    """
    Transform each ``Doc`` or ``Span`` in a sequence thereof into a sequence of terms
    as strings, where "terms" are the concatenated combination of constituent
    n-grams, entities, and/or noun chunks.

    Args:
        ngrams
        entities
        noun_chunks
        normalize
        dedupe
    """

    def __init__(
        self,
        *,
        ngrams: Optional[int | Collection[int] | DocLikeToSpans] = None,
        entities: Optional[bool | DocLikeToSpans] = None,
        noun_chunks: Optional[bool | DocLikeToSpans] = None,
        normalize: Optional[str | Callable[[Span], str]] = None,
        dedupe: bool = True,
    ):
        self.tokenizers = self._init_tokenizers(ngrams, entities, noun_chunks)
        self.normalize = self._init_normalize(normalize)
        self.dedupe = dedupe

    def __str__(self) -> str:
        return (
            "TermsTokenizer("
            f"tokenizers={self.tokenizers}, normalize={self.normalize}, dedupe={self.dedupe}"
            ")"
        )

    def _init_tokenizers(
        self, ngrams, entities, noun_chunks
    ) -> Tuple[DocLikeToSpans, ...]:
        ngs_tokenizer = self._init_ngrams_tokenizer(ngrams)
        ents_tokenizer = self._init_entities_tokenizer(entities)
        ncs_tokenizer = self._init_noun_chunks_tokenizer(noun_chunks)
        tokenizers = tuple(
            tokenizer
            for tokenizer in [ngs_tokenizer, ents_tokenizer, ncs_tokenizer]
            if tokenizer is not None
        )
        if not tokenizers:
            raise ValueError("at least one term tokenizer must be specified")
        else:
            return tokenizers

    def _init_ngrams_tokenizer(
        self, ngrams: Optional[int | Collection[int] | DocLikeToSpans]
    ) -> Optional[Callable[[types.DocLike], Iterable[Span]]]:
        if ngrams is None:
            return None
        elif callable(ngrams):
            return ngrams
        elif isinstance(ngrams, int):
            return partial(extract.ngrams, n=ngrams)
        elif (
            isinstance(ngrams, Collection)
            and all(isinstance(ng, int) for ng in ngrams)
        ):
            return partial(_concat_extract_ngrams, ns=ngrams)
        else:
            raise TypeError()

    def _init_entities_tokenizer(
        self, entities: Optional[bool | DocLikeToSpans]
    ) -> Optional[DocLikeToSpans]:
        if entities is None:
            return None
        elif callable(entities):
            return entities
        elif isinstance(entities, bool):
            return extract.entities
        else:
            raise TypeError()

    def _init_noun_chunks_tokenizer(
        self,
        noun_chunks: Optional[bool | DocLikeToSpans],
    ) -> Optional[DocLikeToSpans]:
        if noun_chunks is None:
            return None
        elif callable(noun_chunks):
            return noun_chunks
        elif isinstance(noun_chunks, bool):
            return extract.noun_chunks
        else:
            raise TypeError()

    def _init_normalize(
        self, normalize: Optional[str | Callable[[Span], str]]
    ) -> Callable[[Span], str]:
        if not normalize:
            return operator.attrgetter("text")
        elif normalize == "lower":
            return lambda span: span.text.lower()
        elif normalize == "lemma":
            return lambda span: span.lemma_
        elif callable(normalize):
            return normalize
        else:
            raise ValueError()

    def fit(self, doclikes: Iterable[types.DocLike]) -> "TermsTokenizer":
        return self

    def transform(self, doclikes: Iterable[types.DocLike]) -> Iterable[Tuple[str, ...]]:
        """
        Convert a sequence of spaCy Docs or Spans into an ordered, nested sequence
        of terms as strings.

        Args:
            doclikes

        Yields:
            Ordered sequence of terms as strings for next Doc or Span.
        """
        normalize_ = self.normalize
        for doclike in doclikes:
            terms = itertoolz.concat(tokenizer(doclike) for tokenizer in self.tokenizers)
            if self.dedupe is True:
                terms = itertoolz.unique(terms, lambda span: (span.start, span.end))
            yield tuple(normalize_(term) for term in terms)


def _concat_extract_ngrams(doclike: types.DocLike, ns: Collection[int]) -> Iterable[Span]:
    for n in ns:
        ngrams = extract.ngrams(doclike, n=n)
        for ngram in ngrams:
            yield ngram


# TODO: do we want to try lazy imports here?
# import importlib
# def __getattr__(name: str) -> Any:
#     if name == "extract":
#         return importlib.import_module(f".{name}", package="textacy")



"""
Basics
------

:mod:`textacy.extract.basics`: Extract basic components from a document or sentence
via spaCy, with bells and whistles for filtering the results.
"""
from __future__ import annotations

from functools import partial
from typing import Collection, Iterable, List, Optional, Set, Union

from cytoolz import itertoolz
from spacy.parts_of_speech import DET
from spacy.tokens import Span, Token

from .. import constants, errors, types, utils


def words(
    doclike: types.DocLike,
    *,
    filter_stops: bool = True,
    filter_punct: bool = True,
    filter_nums: bool = False,
    include_pos: Optional[str | Collection[str]] = None,
    exclude_pos: Optional[str | Collection[str]] = None,
    min_freq: int = 1,
) -> Iterable[Token]:
    """
    Extract an ordered sequence of words from a document processed by spaCy,
    optionally filtering words by part-of-speech tag and frequency.

    Args:
        doclike
        filter_stops: If True, remove stop words from word list.
        filter_punct: If True, remove punctuation from word list.
        filter_nums: If True, remove number-like words (e.g. 10, "ten")
            from word list.
        include_pos: Remove words whose part-of-speech tag IS NOT in the specified tags.
        exclude_pos: Remove words whose part-of-speech tag IS in the specified tags.
        min_freq: Remove words that occur in ``doclike`` fewer than ``min_freq`` times.

    Yields:
        Next token from ``doclike`` passing specified filters in order of appearance
        in the document.

    Raises:
        TypeError: if ``include_pos`` or ``exclude_pos`` is not a str, a set of str,
            or a falsy value

    Note:
        Filtering by part-of-speech tag uses the universal POS tag set; for details,
        check spaCy's docs: https://spacy.io/api/annotation#pos-tagging
    """
    words_: Iterable[Token] = (w for w in doclike if not w.is_space)
    if filter_stops is True:
        words_ = (w for w in words_ if not w.is_stop)
    if filter_punct is True:
        words_ = (w for w in words_ if not w.is_punct)
    if filter_nums is True:
        words_ = (w for w in words_ if not w.like_num)
    if include_pos:
        include_pos = utils.to_collection(include_pos, str, set)
        include_pos = {pos.upper() for pos in include_pos}
        words_ = (w for w in words_ if w.pos_ in include_pos)
    if exclude_pos:
        exclude_pos = utils.to_collection(exclude_pos, str, set)
        exclude_pos = {pos.upper() for pos in exclude_pos}
        words_ = (w for w in words_ if w.pos_ not in exclude_pos)
    if min_freq > 1:
        words_ = list(words_)
        freqs = itertoolz.frequencies(w.lower_ for w in words_)
        words_ = (w for w in words_ if freqs[w.lower_] >= min_freq)

    for word in words_:
        yield word


def ngrams(
    doclike: types.DocLike,
    n: int | Collection[int],
    *,
    filter_stops: bool = True,
    filter_punct: bool = True,
    filter_nums: bool = False,
    include_pos: Optional[str | Collection[str]] = None,
    exclude_pos: Optional[str | Collection[str]] = None,
    min_freq: int = 1,
) -> Iterable[Span]:
    """
    Extract an ordered sequence of n-grams (``n`` consecutive tokens) from a spaCy
    ``Doc`` or ``Span``, for one or multiple ``n`` values, optionally filtering n-grams
    by the types and parts-of-speech of the constituent tokens.

    Args:
        doclike
        n: Number of tokens included per n-gram; for example, ``2`` yields bigrams
            and ``3`` yields trigrams. If multiple values are specified, then the
            collections of n-grams are concatenated together; for example, ``(2, 3)``
            yields bigrams and then trigrams.
        filter_stops: If True, remove ngrams that start or end with a stop word.
        filter_punct: If True, remove ngrams that contain any punctuation-only tokens.
        filter_nums: If True, remove ngrams that contain any numbers
            or number-like tokens (e.g. 10, 'ten').
        include_pos: Remove ngrams if any constituent tokens' part-of-speech tags
            ARE NOT included in this param.
        exclude_pos: Remove ngrams if any constituent tokens' part-of-speech tags
            ARE included in this param.
        min_freq: Remove ngrams that occur in ``doclike`` fewer than ``min_freq`` times

    Yields:
        Next ngram from ``doclike`` passing all specified filters, in order of appearance
        in the document.

    Raises:
        ValueError: if any ``n`` < 1
        TypeError: if ``include_pos`` or ``exclude_pos`` is not a str, a set of str,
            or a falsy value

    Note:
        Filtering by part-of-speech tag uses the universal POS tag set; for details,
        check spaCy's docs: https://spacy.io/api/annotation#pos-tagging
    """
    ns = utils.to_collection(n, int, tuple)
    if any(n_ < 1 for n_ in ns):
        raise ValueError("n must be greater than or equal to 1")

    if include_pos:
        include_pos = {
            pos.upper() for pos in utils.to_collection(include_pos, str, set)
        }
    if exclude_pos:
        exclude_pos = {
            pos.upper() for pos in utils.to_collection(exclude_pos, str, set)
        }
    for n_ in ns:
        ngrams_ = (doclike[i : i + n_] for i in range(len(doclike) - n_ + 1))
        ngrams_ = (ng for ng in ngrams_ if not any(w.is_space for w in ng))
        if filter_stops is True:
            ngrams_ = (ng for ng in ngrams_ if not ng[0].is_stop and not ng[-1].is_stop)
        if filter_punct is True:
            ngrams_ = (ng for ng in ngrams_ if not any(w.is_punct for w in ng))
        if filter_nums is True:
            ngrams_ = (ng for ng in ngrams_ if not any(w.like_num for w in ng))
        if include_pos:
            ngrams_ = (ng for ng in ngrams_ if all(w.pos_ in include_pos for w in ng))
        if exclude_pos:
            ngrams_ = (ng for ng in ngrams_ if not any(w.pos_ in exclude_pos for w in ng))
        if min_freq > 1:
            ngrams_ = list(ngrams_)
            freqs = itertoolz.frequencies(ng.text.lower() for ng in ngrams_)
            ngrams_ = (ng for ng in ngrams_ if freqs[ng.text.lower()] >= min_freq)

        for ngram in ngrams_:
            yield ngram


def entities(
    doclike: types.DocLike,
    *,
    include_types: Optional[str | Collection[str]] = None,
    exclude_types: Optional[str | Collection[str]] = None,
    drop_determiners: bool = True,
    min_freq: int = 1,
) -> Iterable[Span]:
    """
    Extract an ordered sequence of named entities (PERSON, ORG, LOC, etc.) from
    a ``Doc``, optionally filtering by entity types and frequencies.

    Args:
        doclike
        include_types: Remove entities whose type IS NOT
            in this param; if "NUMERIC", all numeric entity types ("DATE",
            "MONEY", "ORDINAL", etc.) are included
        exclude_types: Remove entities whose type IS
            in this param; if "NUMERIC", all numeric entity types ("DATE",
            "MONEY", "ORDINAL", etc.) are excluded
        drop_determiners: Remove leading determiners (e.g. "the")
            from entities (e.g. "the United States" => "United States").

            .. note:: Entities from which a leading determiner has been removed
               are, effectively, *new* entities, and not saved to the ``Doc``
               from which they came. This is irritating but unavoidable, since
               this function is not meant to have side-effects on document state.
               If you're only using the text of the returned spans, this is no
               big deal, but watch out if you're counting on determiner-less
               entities associated with the doc downstream.

        min_freq: Remove entities that occur in ``doclike`` fewer
            than ``min_freq`` times

    Yields:
        Next entity from ``doclike`` passing all specified filters in order of appearance
        in the document

    Raises:
        TypeError: if ``include_types`` or ``exclude_types`` is not a str, a set of
            str, or a falsy value
    """
    ents = doclike.ents
    # HACK: spacy's models have been erroneously tagging whitespace as entities
    # https://github.com/explosion/spaCy/commit/1e6725e9b734862e61081a916baf440697b9971e
    ents = (ent for ent in ents if not ent.text.isspace())
    include_types = _parse_ent_types(include_types, "include")
    exclude_types = _parse_ent_types(exclude_types, "exclude")
    if include_types:
        if isinstance(include_types, str):
            ents = (ent for ent in ents if ent.label_ == include_types)
        elif isinstance(include_types, (set, frozenset, list, tuple)):
            ents = (ent for ent in ents if ent.label_ in include_types)
    if exclude_types:
        if isinstance(exclude_types, str):
            ents = (ent for ent in ents if ent.label_ != exclude_types)
        elif isinstance(exclude_types, (set, frozenset, list, tuple)):
            ents = (ent for ent in ents if ent.label_ not in exclude_types)
    if drop_determiners is True:
        ents = (
            ent
            if ent[0].pos != DET
            else Span(
                ent.doc, ent.start + 1, ent.end, label=ent.label, vector=ent.vector
            )
            for ent in ents
        )
    if min_freq > 1:
        ents = list(ents)
        freqs = itertoolz.frequencies(ent.text.lower() for ent in ents)
        ents = (ent for ent in ents if freqs[ent.text.lower()] >= min_freq)

    for ent in ents:
        yield ent


def _parse_ent_types(
    ent_types: Optional[str | Collection[str]], which: str,
) -> Optional[str | Set[str]]:
    if not ent_types:
        return None
    elif isinstance(ent_types, str):
        ent_types = ent_types.upper()
        # replace the shorthand numeric case by its corresponding constant
        if ent_types == "NUMERIC":
            return constants.NUMERIC_ENT_TYPES
        else:
            return ent_types
    elif isinstance(ent_types, (set, frozenset, list, tuple)):
        ent_types = {ent_type.upper() for ent_type in ent_types}
        # again, replace the shorthand numeric case by its corresponding constant
        # and include it in the set in case other types are specified
        if any(ent_type == "NUMERIC" for ent_type in ent_types):
            return ent_types.union(constants.NUMERIC_ENT_TYPES)
        else:
            return ent_types
    else:
        raise TypeError(
            errors.type_invalid_msg(
                f"{which}_types", type(ent_types), Optional[Union[str, Collection[str]]]
            )
        )


def noun_chunks(
    doclike: types.DocLike, *, drop_determiners: bool = True, min_freq: int = 1,
) -> Iterable[Span]:
    """
    Extract an ordered sequence of noun chunks from a spacy-parsed doc, optionally
    filtering by frequency and dropping leading determiners.

    Args:
        doclike
        drop_determiners: Remove leading determiners (e.g. "the")
            from phrases (e.g. "the quick brown fox" => "quick brown fox")
        min_freq: Remove chunks that occur in ``doclike`` fewer than ``min_freq`` times

    Yields:
        Next noun chunk from ``doclike`` in order of appearance in the document
    """
    ncs = doclike.noun_chunks
    if drop_determiners is True:
        ncs = (nc if nc[0].pos != DET else nc[1:] for nc in ncs)
    if min_freq > 1:
        ncs = list(ncs)
        freqs = itertoolz.frequencies(nc.text.lower() for nc in ncs)
        ncs = (nc for nc in ncs if freqs[nc.text.lower()] >= min_freq)

    for nc in ncs:
        yield nc


def terms(
    doclike: types.DocLike,
    *,
    ngs: Optional[int | Collection[int] | types.DocLikeToSpans] = None,
    ents: Optional[bool | types.DocLikeToSpans] = None,
    ncs: Optional[bool | types.DocLikeToSpans] = None,
    dedupe: bool = True,
) -> Iterable[Span]:
    """
    Extract one or multiple types of terms -- ngrams, entities, and/or noun chunks --
    from ``doclike`` as a single, concatenated collection, with optional deduplication
    of spans extracted by more than one type.

    .. code-block:: pycon

        >>> extract.terms(doc, ngs=2, ents=True, ncs=True)
        >>> extract.terms(doc, ngs=lambda doc: extract.ngrams(doc, n=2))
        >>> extract.terms(doc, ents=extract.entities)
        >>> extract.terms(doc, ents=partial(extract.entities, include_types="PERSON"))

    Args:
        doclike
        ngs: N-gram terms to be extracted.
            If one or multiple ints, :func:`textacy.extract.ngrams(doclike, n=ngs)` is
            used to extract terms; if a callable, ``ngs(doclike)`` is used to extract
            terms; if None, no n-gram terms are extracted.
        ents: Entity terms to be extracted.
            If True, :func:`textacy.extract.entities(doclike)` is used to extract terms;
            if a callable, ``ents(doclike)`` is used to extract terms;
            if None, no entity terms are extracted.
        ncs: Noun chunk terms to be extracted.
            If True, :func:`textacy.extract.noun_chunks(doclike)` is used to extract
            terms; if a callable, ``ncs(doclike)`` is used to extract terms;
            if None, no noun chunk terms are extracted.
        dedupe: If True, deduplicate terms whose spans are extracted by multiple types
            (e.g. a span that is both an n-gram and an entity), as identified by
            identical (start, stop) indexes in ``doclike``; otherwise, don't.

    Returns:
        Next term from ``doclike``, in order of n-grams then entities then noun chunks,
        with each collection's terms given in order of appearance.

    Note:
        This function is *not* to be confused with keyterm extraction, which leverages
        statistics and algorithms to quantify the "key"-ness of terms before returning
        the top-ranking terms. There is no such scoring or ranking here.

    See Also:
        - :func:`textacy.extact.ngrams()`
        - :func:`textacy.extact.entities()`
        - :func:`textacy.extact.noun_chunks()`
        - :mod:`textacy.extact.keyterms`
    """
    extractors = _get_extractors(ngs, ents, ncs)
    terms_ = itertoolz.concat(extractor(doclike) for extractor in extractors)
    if dedupe is True:
        terms_ = itertoolz.unique(terms_, lambda span: (span.start, span.end))
    for term in terms_:
        yield term


def _get_extractors(ngs, ents, ncs) -> List[types.DocLikeToSpans]:
    all_extractors = [
        _get_ngs_extractor(ngs), _get_ents_extractor(ents), _get_ncs_extractor(ncs)
    ]
    extractors = [extractor for extractor in all_extractors if extractor is not None]
    if not extractors:
        raise ValueError("at least one term extractor must be specified")
    else:
        return extractors


def _get_ngs_extractor(ngs) -> Optional[types.DocLikeToSpans]:
    if ngs is None:
        return None
    elif callable(ngs):
        return ngs
    elif (
        isinstance(ngs, int)
        or (isinstance(ngs, Collection) and all(isinstance(ng, int) for ng in ngs))
    ):
        return partial(ngrams, n=ngs)
    else:
        raise TypeError()


def _get_ents_extractor(ents) -> Optional[types.DocLikeToSpans]:
    if ents is None:
        return None
    elif callable(ents):
        return ents
    elif isinstance(ents, bool):
        return entities
    else:
        raise TypeError()


def _get_ncs_extractor(ncs) -> Optional[types.DocLikeToSpans]:
    if ncs is None:
        return None
    elif callable(ncs):
        return ncs
    elif isinstance(ncs, bool):
        return noun_chunks
    else:
        raise TypeError()



import pytest

import textacy
from textacy import datasets
from textacy.extract import keyterms as kt


DATASET = datasets.CapitolWords()

pytestmark = pytest.mark.skipif(
    DATASET.filepath is None,
    reason="CapitolWords dataset must be downloaded before running tests",
)


@pytest.fixture(scope="module")
def spacy_doc():
    text = next(DATASET.texts(min_len=1500, limit=1))
    return textacy.make_spacy_doc(text, lang="en_core_web_sm")


@pytest.fixture(scope="module")
def empty_spacy_doc():
    return textacy.make_spacy_doc("", lang="en_core_web_sm")


def test_default(spacy_doc):
    result = kt.scake(spacy_doc)
    assert isinstance(result, list) and len(result) > 0
    assert all(isinstance(ts, tuple) and len(ts) == 2 for ts in result)
    assert all(
        isinstance(ts[0], str) and isinstance(ts[1], float)
        for ts in result
    )


def test_include_pos(spacy_doc):
    result1 = kt.scake(spacy_doc, include_pos={"NOUN", "PROPN", "ADJ"})
    result2 = kt.scake(spacy_doc, include_pos={"NOUN", "PROPN"})
    assert len(result1) > 0 and len(result2) > 0
    assert result1 != result2


def test_n_topn(spacy_doc):
    for n in (5, 25):
        result = kt.scake(spacy_doc, topn=n)
        assert 0 < len(result) <= n


def test_topn_float(spacy_doc):
    result = kt.scake(spacy_doc, topn=0.2)
    assert len(result) > 0
    with pytest.raises(ValueError):
        _ = kt.scake(spacy_doc, topn=2.0)


def test_empty_doc(empty_spacy_doc):
    result = kt.scake(empty_spacy_doc)
    assert isinstance(result, list)
    assert len(result) == 0


def test_single_sentence_doc():
    doc = textacy.make_spacy_doc(
        "This is a document with a single sentence.",
        lang="en_core_web_sm",
    )
    result = kt.scake(doc)
    assert isinstance(result, list)
    assert len(result) > 0



import datetime
import os

import pytest

from textacy.datasets.capitol_words import CapitolWords

DATASET = CapitolWords()

pytestmark = pytest.mark.skipif(
    DATASET.filepath is None,
    reason="CapitolWords dataset must be downloaded before running tests",
)


@pytest.mark.skip("No need to download a new dataset every time")
def test_download(tmpdir):
    dataset = CapitolWords(data_dir=str(tmpdir))
    dataset.download()
    assert os.path.isfile(dataset._filepath)


def test_oserror(tmpdir):
    dataset = CapitolWords(data_dir=str(tmpdir))
    with pytest.raises(OSError):
        _ = list(dataset.texts())


def test_texts():
    for text in DATASET.texts(limit=3):
        assert isinstance(text, str)


def test_texts_limit():
    for limit in (1, 5, 10):
        assert sum(1 for _ in DATASET.texts(limit=limit)) == limit


def test_texts_min_len():
    for min_len in (100, 200, 1000):
        assert all(
            len(text) >= min_len for text in DATASET.texts(min_len=min_len, limit=10)
        )


def test_records():
    for text, meta in DATASET.records(limit=3):
        assert isinstance(text, str)
        assert isinstance(meta, dict)


def test_records_speaker_name():
    speaker_names = ({"Bernie Sanders"}, {"Joseph Biden", "Rick Santorum"})
    for speaker_name in speaker_names:
        records = list(DATASET.records(speaker_name=speaker_name, limit=3))
        assert len(records) >= 1
        assert all(
            meta["speaker_name"] in speaker_name
            for text, meta in records
        )


def test_records_speaker_party():
    speaker_parties = ({"R"}, {"D", "I"})
    for speaker_party in speaker_parties:
        records = list(DATASET.records(speaker_party=speaker_party, limit=10))
        assert len(records) >= 1
        assert all(
            meta["speaker_party"] in speaker_party
            for text, meta in records
        )


def test_records_chamber():
    chambers = ({"House"}, {"House", "Senate"})
    for chamber in chambers:
        records = list(DATASET.records(chamber=chamber, limit=10))
        assert len(records) >= 1
        assert all(
            meta["chamber"] in chamber
            for text, meta in records
        )


def test_records_congress():
    congresses = ({104}, {104, 114})
    for congress in congresses:
        records = list(DATASET.records(congress=congress, limit=10))
        assert len(records) >= 1
        assert all(
            meta["congress"] in congress
            for text, meta in records
        )


def test_records_date_range():
    date_ranges = (["1997-01-01", "1998-01-01"], ("1997-01-01", "1997-02-01"))
    for date_range in date_ranges:
        records = list(DATASET.records(date_range=date_range, limit=3))
        assert len(records) >= 1
        assert all(
            date_range[0] <= meta["date"] < date_range[1]
            for text, meta in records
        )


def test_bad_filters():
    bad_filters = (
        {"speaker_name": "Burton DeWilde"},
        {"speaker_party": "Whigs"},
        {"chamber": "White House"},
        {"congress": 42},
        {"min_len": -1},
    )
    for bad_filter in bad_filters:
        with pytest.raises(ValueError):
            list(DATASET.texts(**bad_filter))
    bad_filters = (
        {"score_range": ["low", "high"]},
        {"date_range": "2016-01-01"},
        {"date_range": (datetime.date(2000, 1, 1), datetime.date(2001, 1, 1))},
    )
    for bad_filter in bad_filters:
        with pytest.raises(TypeError):
            list(DATASET.texts(**bad_filter))
